# Generated by Django 3.1.3 on 2020-11-30 10:40

import os
import uuid

from django.db import migrations, models


def migrate_data_sources(apps, schema_editor):
    DataSource = apps.get_model('documents', 'DataSource')
    if not DataSource.objects.exists():
        return

    repo_prefix = os.environ.get('REPOSITORY_PREFIX', 'https://gitlab.com/teklia/workers')

    Element = apps.get_model('documents', 'Element')
    Classification = apps.get_model('documents', 'Classification')
    Transcription = apps.get_model('documents', 'Transcription')
    Entity = apps.get_model('documents', 'Entity')
    Repository = apps.get_model('process', 'Repository')
    Worker = apps.get_model('process', 'Worker')

    # Start by just removing the manual source, since a manual worker version is just None
    print('Migrating manual sources…')
    Element.objects.filter(source__slug='manual').update(source=None)
    Classification.objects.filter(source__slug='manual').update(source=None)
    Transcription.objects.filter(source__slug='manual').update(source=None)
    Entity.objects.filter(source__slug='manual').update(source=None)

    # Only migrate sources that have related objects
    to_migrate = DataSource.objects.filter(
        id__in=Element.objects.values('source_id').union(
            Classification.objects.values('source_id')
        ).union(
            Transcription.objects.values('source_id')
        ).union(
            Entity.objects.values('source_id')
        )
    )
    for source in to_migrate:
        print(f'Migrating {source.name} {source.revision} ({source.id})…')
        repo, _ = Repository.objects.get_or_create(
            url=f'{repo_prefix}/{source.slug}',
            type='worker',
            defaults={
                # This is supposed to be unique
                'hook_token': str(source.id),
            },
        )
        worker, _ = Worker.objects.get_or_create(
            repository=repo,
            slug=source.slug,
            defaults={
                'name': source.name,
                'type': str(source.type),
            }
        )

        # To ensure we cannot get duplicate worker version IDs when re-applying
        # unique constraints later, we just keep on trying to get a new revision.
        created = False
        while not created:
            revision, created = repo.revisions.get_or_create(
                hash=uuid.uuid4().hex,
                message='Migrated DataSource',
                author='Arkindex',
            )

        version = worker.workerversion_set.create(
            revision=revision,
            configuration={},
        )

        source.elements.update(source=None, worker_version=version)
        source.classifications.update(source=None, worker_version=version)
        source.transcriptions.update(source=None, worker_version=version)
        source.entities.update(source=None, worker_version=version)


class Migration(migrations.Migration):

    dependencies = [
        ('documents', '0023_remove_transcription_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='DataSource',
            name='type',
            field=models.CharField(max_length=50),
        ),
        migrations.RemoveConstraint(
            model_name='classification',
            name='classification_unique_manual',
        ),
        migrations.RemoveConstraint(
            model_name='classification',
            name='classification_unique_worker_version',
        ),
        migrations.RemoveConstraint(
            model_name='transcription',
            name='transcription_source_not_worker_version',
        ),
        migrations.RunPython(
            code=migrate_data_sources,
            reverse_code=migrations.RunPython.noop,
            elidable=True,
        ),
        # Duplicate classifications may still remain
        migrations.RunSQL(
            """
            DELETE FROM documents_classification WHERE id IN (
                SELECT id
                FROM (
                    SELECT c.id, ROW_NUMBER() OVER (
                        PARTITION BY x.element_id
                        ORDER BY c.confidence DESC
                    ) AS nb
                    FROM (
                        SELECT element_id, ml_class_id
                        FROM documents_classification
                        GROUP BY element_id, ml_class_id
                        HAVING COUNT(id) > 1
                      ) AS x
                    INNER JOIN documents_classification AS c USING (element_id, ml_class_id)
                ) AS y
                WHERE y.nb > 1
            );
            """,
            reverse_sql=migrations.RunSQL.noop,
            elidable=True,
        ),
        # Deletion happens in another migration, since updating data then trying to update
        # the structure causes errors with 'pending trigger events'
    ]
