import uuid

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestBulkTranscriptionEntities(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_run = cls.worker_version.worker_runs.get()
        cls.transcription = cls.corpus.elements.get(name='Volume 1, page 1r').transcriptions.create(
            text='Once upon a time in a castle a knight was living lonely',
            confidence=0.42,
        )
        cls.location_ent_type = cls.corpus.entity_types.get(name="location")
        cls.person_ent_type = cls.corpus.entity_types.get(name="person")
        cls.entity = cls.corpus.entities.create(name="Paris", type=cls.location_ent_type)

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:transcription-entities-bulk', kwargs={'pk': self.transcription.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:transcription-entities-bulk', kwargs={'pk': self.transcription.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_requires_contributor(self):
        self.user.rights.update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:transcription-entities-bulk', kwargs={'pk': self.transcription.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {
            'detail': 'You do not have a contributor access to the corpus of this transcription.'
        })

    def test_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse('api:transcription-entities-bulk', kwargs={'pk': uuid.uuid4()}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_required_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:transcription-entities-bulk', kwargs={'pk': str(self.transcription.id)}),
                data={},
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'entities': ['This field is required.'],
            'worker_run_id': ['This field is required.'],
        })

    def test_wrong_values(self):
        fake_uuid = uuid.uuid4()
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:transcription-entities-bulk', kwargs={'pk': str(self.transcription.id)}),
                data={
                    'entities': [],
                    'worker_run_id': str(fake_uuid),
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'entities': {'non_field_errors': ['This list may not be empty.']},
            'worker_run_id': [f'Invalid pk "{fake_uuid}" - object does not exist.'],
        })

    def test_existing_transcription_entity(self):
        self.transcription.transcription_entities.create(
            entity=self.entity,
            offset=1,
            length=1,
            worker_run=self.worker_run,
            worker_version=self.worker_run.version,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:transcription-entities-bulk', kwargs={'pk': str(self.transcription.id)}),
                data={
                    'entities': [
                        {
                            'name': 'Knight',
                            'type_id': str(self.person_ent_type.id),
                            'offset': 10,
                            'length': 5,
                            'confidence': 0.05,
                        },
                    ],
                    'worker_run_id': str(self.worker_run.id),
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_run_id': ['Some TranscriptionEntities were already created for this worker run on this transcription.'],
        })

    def test_entity_fields_validation(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:transcription-entities-bulk', kwargs={'pk': str(self.transcription.id)}),
                data={
                    'entities': [
                        {}
                    ],
                    'worker_run_id': str(self.worker_run.id),
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'entities': [{
                'confidence': ['This field is required.'],
                'length': ['This field is required.'],
                'name': ['This field is required.'],
                'offset': ['This field is required.'],
                'type_id': ['This field is required.'],
            }]
        })

    def test_duplicate_entities(self):
        self.client.force_login(self.user)
        payload = {
            'name': 'Knight',
            'type_id': str(self.person_ent_type.id),
            'offset': 10,
            'length': 5,
            'confidence': 0.05,
        }
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:transcription-entities-bulk', kwargs={'pk': str(self.transcription.id)}),
                data={
                    'entities': [
                        payload,
                        payload,
                    ],
                    'worker_run_id': str(self.worker_run.id),
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'entities': {
                '__all__': ['Some TranscriptionEntities have the same name, type, offset and length.'],
            },
        })

    def test_unknown_type(self):
        pass

    def test_new_entities(self):
        self.client.force_login(self.user)
        self.assertQuerysetEqual(
            self.corpus.entities.values_list('name', 'type', 'worker_run_id').order_by('name', 'type', 'worker_run_id'),
            [('Paris', self.location_ent_type.id, None)],
        )
        self.assertEqual(self.transcription.transcription_entities.count(), 0)

        with self.assertNumQueries(13):
            response = self.client.post(
                reverse('api:transcription-entities-bulk', kwargs={'pk': str(self.transcription.id)}),
                data={
                    'entities': [
                        {
                            # This entity already exists, and should be duplicated
                            'name': 'Paris',
                            'type_id': str(self.location_ent_type.id),
                            'offset': 0,
                            'length': 1,
                            'confidence': 0.7,
                        },
                        {
                            'name': 'Knight',
                            'type_id': str(self.person_ent_type.id),
                            'offset': 0,
                            'length': 1,
                            'confidence': 0.7,
                        },
                        {
                            'name': 'Knight',
                            'type_id': str(self.person_ent_type.id),
                            'offset': 10,
                            'length': 5,
                            'confidence': 0.05,
                        },
                    ],
                    'worker_run_id': str(self.worker_run.id),
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertQuerysetEqual(
            self.corpus.entities.values_list('name', 'type', 'worker_run_id').order_by('name', 'type', 'worker_run_id'),
            [
                ('Knight', self.person_ent_type.id, self.worker_run.id),
                ('Knight', self.person_ent_type.id, self.worker_run.id),
                ('Paris', self.location_ent_type.id, self.worker_run.id),
                ('Paris', self.location_ent_type.id, None),
            ],
        )

        self.assertQuerysetEqual(
            self.transcription.transcription_entities
            .values_list('entity__name', 'entity__type', 'offset', 'length', 'confidence', 'worker_run_id')
            .order_by('entity__name', 'offset'),
            [
                ('Knight', self.person_ent_type.id, 0, 1, 0.7, self.worker_run.id),
                ('Knight', self.person_ent_type.id, 10, 5, 0.05, self.worker_run.id),
                ('Paris', self.location_ent_type.id, 0, 1, 0.7, self.worker_run.id),
            ],
        )

        knight1, knight2, paris = self.transcription.transcription_entities.order_by('entity__name', 'offset')

        # Ensure that the returned IDs match the order in which the payloads were sent
        self.assertEqual(response.json(), {
            "entities": [
                {
                    'transcription_entity_id': paris.id,
                    'entity_id': str(paris.entity_id),
                },
                {
                    'transcription_entity_id': knight1.id,
                    'entity_id': str(knight1.entity_id),
                },
                {
                    'transcription_entity_id': knight2.id,
                    'entity_id': str(knight2.entity_id),
                },
            ],
        })
