from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import TextOrientation
from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestBulkTranscriptions(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_run = cls.worker_version.worker_runs.get()

    def test_bulk_transcriptions_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:transcription-bulk'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_bulk_transcriptions_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:transcription-bulk'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_bulk_transcriptions_not_found(self):
        self.client.force_login(self.user)
        self.user.rights.all().delete()
        forbidden_element = self.corpus.elements.get(name='Volume 1, page 1r')
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:transcription-bulk'), {
                "worker_run_id": str(self.worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(forbidden_element.id),
                        "text": "lol",
                        "confidence": 0.4
                    },
                    {
                        "element_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                        "text": "lol",
                        "confidence": 0.4
                    }
                ],
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'transcriptions': [
                {'element_id': [f'Element {forbidden_element.id} was not found or cannot be written to.']},
                {'element_id': ['Element aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa was not found or cannot be written to.']},
            ]
        })

    def test_bulk_transcriptions_worker_version(self):
        self.client.force_login(self.user)

        element1 = self.corpus.elements.get(name='Volume 2')
        element2 = self.corpus.elements.get(name='Volume 2, page 1r')
        self.assertFalse(element1.transcriptions.exists())
        self.assertFalse(element2.transcriptions.exists())

        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:transcription-bulk'), {
                "worker_version": str(self.worker_version.id),
                "transcriptions": [
                    {
                        "element_id": str(element1.id),
                        "text": "Sneasel",
                        "confidence": 0.54
                    },
                    {
                        "element_id": str(element2.id),
                        "text": "Charizard",
                        "confidence": 0.85
                    },
                    {
                        "element_id": str(element1.id),
                        "text": "Raticate",
                        "confidence": 0.12
                    },
                ],
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_run_id': ['This field is required.'],
            'worker_version': ['This field is forbidden.'],
        })

    def test_bulk_transcriptions_worker_run(self):
        self.client.force_login(self.user)

        element1 = self.corpus.elements.get(name='Volume 2')
        element2 = self.corpus.elements.get(name='Volume 2, page 1r')
        self.assertFalse(element1.transcriptions.exists())
        self.assertFalse(element2.transcriptions.exists())

        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:transcription-bulk'), {
                "worker_run_id": str(self.worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(element1.id),
                        "text": "Sneasel",
                        "confidence": 0.54
                    },
                    {
                        "element_id": str(element2.id),
                        "text": "Charizard",
                        "confidence": 0.85
                    },
                    {
                        "element_id": str(element1.id),
                        "text": "Raticate",
                        "confidence": 0.12
                    },
                ],
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        first_tr, second_tr = element1.transcriptions.values_list('id', flat=True)
        third_tr = element2.transcriptions.values_list('id', flat=True)[0]
        self.assertEqual(response.json(), {
            "worker_run_id": str(self.worker_run.id),
            "transcriptions": [
                {
                    "id": str(first_tr),
                    "element_id": str(element1.id),
                    "text": "Sneasel",
                    "orientation": TextOrientation.HorizontalLeftToRight.value,
                    "confidence": 0.54,
                },
                {
                    "id": str(third_tr),
                    "element_id": str(element2.id),
                    "text": "Charizard",
                    "orientation": TextOrientation.HorizontalLeftToRight.value,
                    "confidence": 0.85,
                },
                {
                    "id": str(second_tr),
                    "element_id": str(element1.id),
                    "text": "Raticate",
                    "orientation": TextOrientation.HorizontalLeftToRight.value,
                    "confidence": 0.12,
                },
            ]
        })

        self.assertCountEqual(
            list(element1.transcriptions.values('text', 'confidence', 'worker_version_id', 'worker_run_id')),
            [
                {
                    "text": "Sneasel",
                    "confidence": 0.54,
                    "worker_version_id": self.worker_version.id,
                    "worker_run_id": self.worker_run.id,
                },
                {
                    "text": "Raticate",
                    "confidence": 0.12,
                    "worker_version_id": self.worker_version.id,
                    "worker_run_id": self.worker_run.id,
                },
            ]
        )
        self.assertCountEqual(
            list(element2.transcriptions.values('text', 'confidence', 'worker_version_id', 'worker_run_id')),
            [{
                "text": "Charizard",
                "confidence": 0.85,
                "worker_version_id": self.worker_version.id,
                "worker_run_id": self.worker_run.id,
            }]
        )

    def test_bulk_transcriptions_requires_confidence(self):
        self.client.force_login(self.user)
        element = self.corpus.elements.first()

        with self.assertNumQueries(3):
            response = self.client.post(reverse('api:transcription-bulk'), {
                "worker_run_id": str(self.worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(element.id),
                        "text": "Sneasel"
                    },
                ],
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'transcriptions': [
            {'non_field_errors': [
                'The confidence field must be defined.'
            ]}
        ]})

    def test_bulk_transcriptions_orientation(self):
        """
        A text orientation can be specified creating transcriptions in bulk
        """
        self.client.force_login(self.user)
        test_element = self.corpus.elements.get(name='Volume 2, page 1r')
        self.assertFalse(test_element.transcriptions.exists())

        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:transcription-bulk'), {
                "worker_run_id": str(self.worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(test_element.id),
                        "text": "All Hail The Glow Cloud",
                        "orientation": TextOrientation.HorizontalRightToLeft.value,
                        "confidence": 0.78
                    },
                    {
                        "element_id": str(test_element.id),
                        "text": "The Glow Cloud does not need to converse with us.",
                        "orientation": TextOrientation.VerticalRightToLeft.value,
                        "confidence": 0.33
                    },
                ],
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        first_tr, second_tr = test_element.transcriptions.values_list('id', flat=True)
        self.assertEqual(response.json(), {
            "worker_run_id": str(self.worker_run.id),
            "transcriptions": [
                {
                    "id": str(first_tr),
                    "element_id": str(test_element.id),
                    "text": "All Hail The Glow Cloud",
                    "orientation": TextOrientation.HorizontalRightToLeft.value,
                    "confidence": 0.78,
                },
                {
                    "id": str(second_tr),
                    "element_id": str(test_element.id),
                    "text": "The Glow Cloud does not need to converse with us.",
                    "orientation": TextOrientation.VerticalRightToLeft.value,
                    "confidence": 0.33,
                },
            ]
        })

    def test_bulk_transcriptions_invalid_orientation(self):
        """
        An invalid text orientation value causes an error
        """
        self.client.force_login(self.user)
        test_element = self.corpus.elements.get(name='Volume 2, page 1r')
        self.assertFalse(test_element.transcriptions.exists())

        with self.assertNumQueries(3):
            response = self.client.post(reverse('api:transcription-bulk'), {
                "worker_run_id": str(self.worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(test_element.id),
                        "text": "All Hail The Glow Cloud",
                        "orientation": "khoshek",
                        "confidence": 0.78
                    },
                    {
                        "element_id": str(test_element.id),
                        "text": "The Glow Cloud does not need to converse with us.",
                        "orientation": TextOrientation.VerticalRightToLeft.value,
                        "confidence": 0.33
                    },
                ],
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'transcriptions': [
                {'orientation': ['Value is not of type TextOrientation']},
                {}
            ]
        })

    def test_bulk_transcriptions_worker_run_and_version(self):
        self.client.force_login(self.user)
        test_element = self.corpus.elements.get(name='Volume 2, page 1r')
        with self.assertNumQueries(3):
            response = self.client.post(reverse('api:transcription-bulk'), {
                "worker_version": str(self.worker_version.id),
                "worker_run_id": str(self.worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(test_element.id),
                        "text": "The Glow Cloud does not need to converse with us.",
                        "orientation": TextOrientation.VerticalRightToLeft.value,
                        "confidence": 0.33
                    },
                ],
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_version': ['This field is forbidden.'],
        })

    def test_bulk_transcriptions_requires_worker_run(self):
        self.client.force_login(self.user)
        test_element = self.corpus.elements.get(name='Volume 2, page 1r')
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:transcription-bulk'), {
                "transcriptions": [
                    {
                        "element_id": str(test_element.id),
                        "text": "The Glow Cloud does not need to converse with us.",
                        "orientation": TextOrientation.VerticalRightToLeft.value,
                        "confidence": 0.33
                    },
                ],
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_run_id': ['This field is required.'],
        })
