from unittest.mock import patch

from django.test import TestCase

from arkindex.documents.date_parser import (
    DATE_FUNCTIONS_TABLE,
    instanciate_date,
    parse_date,
    year,
    year_month,
    year_month_day,
    year_month_str,
    year_year,
)
from arkindex.documents.dates import DateType, InterpretedDate


class TestDateParser(TestCase):

    def _create_date(self, elts):
        return InterpretedDate(
            year=elts[0],
            month=elts[1],
            day=elts[2],
            type=elts[3],
        )

    @classmethod
    def setUpTestData(cls):
        cls.test_dates = {
            '1343': ((1343, None, None, DateType.Exact), ),
            '1326-05': ((1326, 5, None, DateType.Exact), ),
            '1420-01-23': ((1420, 1, 23, DateType.Exact), ),
            '1398 february': ((1398, 2, None, DateType.Exact), ),
            'Août 1300': ((1300, 8, None, DateType.Exact), ),
            '1332-1337': (
                (1332, None, None, DateType.Lower),
                (1337, None, None, DateType.Upper),
            ),
            'march 12 11223344': (),
            '2222-22-22': (),
        }

    def test_parser(self):
        for t in self.test_dates:
            self.assertTupleEqual(
                parse_date(t),
                tuple([self._create_date(elts) for elts in self.test_dates[t]])
            )

    def _wrong_function(self, raw_date):
        return 'not a dict'

    @patch('logging.Logger.warning')
    def test_parser_exception(self, warning_mock):
        """
        Check that a error-raising function do not break table iteration
        """
        functions = (self._wrong_function, ) + DATE_FUNCTIONS_TABLE
        self.assertTupleEqual(parse_date('1343', functions), (InterpretedDate(year=1343, type=DateType.Exact), ))
        self.assertEqual(warning_mock.call_count, 1)

    def test_year(self):
        dates = ('1343', )
        wrong_dates = ('1343-03', )
        for d in dates:
            self.assertIsInstance(year(d), tuple)
        for d in wrong_dates:
            self.assertIsNone(year(d))

    def test_year_month(self):
        dates = ('1343-02', )
        wrong_dates = ('1343-03-26', '1546')
        for d in dates:
            self.assertIsInstance(year_month(d), tuple)
        for d in wrong_dates:
            self.assertIsNone(year_month(d))

    def test_year_month_day(self):
        dates = ('1343-2-5', )
        wrong_dates = ('1343-03', '1088')
        for d in dates:
            self.assertIsInstance(year_month_day(d), tuple)
        for d in wrong_dates:
            self.assertIsNone(year_month_day(d))

    def test_year_month_str(self):
        dates = ('1343 july', '23 october 1245')
        wrong_dates = ('1343-03', 'july')
        for d in dates:
            self.assertIsInstance(year_month_str(d), tuple)
        for d in wrong_dates:
            self.assertIsNone(year_month_str(d))

    def test_year_year(self):
        dates = ('1342-1245', )
        wrong_dates = ('1345-1342', '13-45 1253', '134301344')
        for d in dates:
            self.assertIsInstance(year_year(d), tuple)
        for d in wrong_dates:
            self.assertIsNone(year_month_str(d))

    def test_instanciate_date(self):
        date = instanciate_date({
            'year': 1320,
            'month': 5,
            'type': DateType.Exact
        })
        self.assertIsInstance(date, InterpretedDate)
        self.assertEqual(date.type, DateType.Exact)
        self.assertEqual(
            tuple(date),
            (1320, 5, None)
        )

    def test_instanciate_wrong_date(self):
        with self.assertRaises(AssertionError):
            instanciate_date('not a dict')
        with self.assertRaises(ValueError):
            instanciate_date({'year': 'not an int'})
