from uuid import uuid4

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element, EntityType, TextOrientation, Transcription
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestEditTranscription(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        page = cls.corpus.elements.get(name='Volume 1, page 1r')
        element = Element.objects.get_descending(page.id).filter(type__slug='word', name='PARIS').get()
        cls.line = cls.corpus.elements.filter(type__slug='text_line').first()
        cls.ml_transcription = element.transcriptions.get(text='PARIS')
        cls.worker_version = cls.ml_transcription.worker_version
        private_corpus = Corpus.objects.create(name='Private')
        cls.private_page = private_corpus.elements.create(type=page.type)
        location_type = EntityType.objects.get(name='location', corpus=cls.corpus)
        cls.entity = cls.corpus.entities.create(name='Some entity', type=location_type)

        # Create an user with a read right only on the private corpus
        cls.private_read_user = User.objects.create_user('a@bc.de', 'a')
        cls.private_read_user.verified_email = True
        cls.private_read_user.save()
        private_corpus.memberships.create(user=cls.private_read_user, level=Role.Guest.value)

        # Create an user with a write right only on the corpus
        cls.write_user = User.objects.create_user('bla@bla.bla', 'aa')
        cls.write_user.verified_email = True
        cls.write_user.save()
        cls.corpus.memberships.create(user=cls.write_user, level=Role.Contributor.value)

    def setUp(self):
        super().setUp()
        self.manual_transcription = self.line.transcriptions.create(text='A manual transcription')
        self.private_transcription = self.private_page.transcriptions.create(text='PEPE')

    def test_get(self):
        """
        Allows to retrieve a single transcription with a read right
        """
        self.client.force_login(self.private_read_user)
        response = self.client.get(reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.manual_transcription.id),
            'confidence': None,
            'text': 'A manual transcription',
            'orientation': TextOrientation.HorizontalLeftToRight.value,
            'worker_version_id': None,
            'worker_run': None,
        })

    def test_get_worker_version(self):
        """
        Retrieve a transcription produced by a ML worker
        """
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:transcription-edit', kwargs={'pk': self.ml_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.ml_transcription.id),
            'confidence': 1.0,
            'text': 'PARIS',
            'orientation': TextOrientation.HorizontalLeftToRight.value,
            'worker_version_id': str(self.worker_version.id),
            'worker_run': None,
        })

    def test_get_verified_user(self):
        response = self.client.get(reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_get_private(self):
        user = User.objects.create_user('f@gh.ij', 'b')
        user.verified_email = True
        user.save()
        self.client.force_login(user)
        response = self.client.get(reverse('api:transcription-edit', kwargs={'pk': self.private_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}),
            format='json',
            data={'text': 'Once upon a time in a castle'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data.get('id'), str(self.manual_transcription.id))
        self.assertEqual(data.get('text'), 'Once upon a time in a castle')

    def test_patch_not_manual(self):
        # self.user is the admin of the corpus, he can edit a worker transcription
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': self.ml_transcription.id}),
            format='json',
            data={'text': 'Once upon a time in a castle'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data.get('id'), str(self.ml_transcription.id))
        self.assertEqual(data.get('text'), 'Once upon a time in a castle')

    def test_patch_internal_user(self):
        self.private_read_user.is_internal = True
        self.private_read_user.save()

        self.client.force_login(self.private_read_user)
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': self.ml_transcription.id}),
            format='json',
            data={'text': 'He used to play the Bumbulum'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data.get('id'), str(self.ml_transcription.id))
        self.assertEqual(data.get('text'), 'He used to play the Bumbulum')

    def test_patch_text_only(self):
        """
        Assert only transcription text can be updated
        """
        self.client.force_login(self.user)
        manual_tr_id = self.manual_transcription.id
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': manual_tr_id}),
            format='json',
            data={
                'id': uuid4(),
                'text': 'a knight was living lonely',
                'type': 'page',
                'confidence': .42,
                'zone': {'polygon': [[4, 2], [2, 4], [3, 3]], 'center': [42, 1337]},
                'image': {'status': 'checked'},
                'element': 'water'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.manual_transcription.refresh_from_db()
        self.assertDictEqual(response.json(), {
            'id': str(manual_tr_id),
            'confidence': None,
            'text': 'a knight was living lonely',
            'orientation': TextOrientation.HorizontalLeftToRight.value,
            'worker_version_id': None,
            'worker_run': None,
        })

    def test_patch_orientation(self):
        """
        Assert it is possible to edit only the text orientation
        """
        self.client.force_login(self.user)
        manual_tr_id = self.manual_transcription.id
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': manual_tr_id}),
            format='json',
            data={
                'orientation': 'vertical-rl'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.manual_transcription.refresh_from_db()
        self.assertDictEqual(response.json(), {
            'id': str(manual_tr_id),
            'confidence': None,
            'text': 'A manual transcription',
            'orientation': 'vertical-rl',
            'worker_version_id': None,
            'worker_run': None,
        })

    def test_patch_invalid_orientation(self):
        """
        An invalid text orientation value causes an error
        """
        self.client.force_login(self.user)
        manual_tr_id = self.manual_transcription.id
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': manual_tr_id}),
            format='json',
            data={
                'orientation': 'wobbly'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'orientation': ['Value is not of type TextOrientation']})

    def test_patch_write_right(self):
        """
        A write right is required to patch a manual transcription
        """
        self.client.force_login(self.private_read_user)
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': self.private_transcription.id}),
            format='json',
            data={'text': 'He was good and fearless'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'A write access to transcription element corpus is required.'
        })

    def test_patch_admin_right(self):
        """
        Updating a transcription produced by a ML worker is forbidden
        """
        self.client.force_login(self.write_user)
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': self.ml_transcription.id}),
            format='json',
            data={'text': 'He used to play the Bumbulum'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Only admins can edit non-manual transcription.'
        })

    def test_patch_entities(self):
        """
        Updating a transcription that has entities is forbidden
        """
        self.client.force_login(self.user)
        self.manual_transcription.transcription_entities.create(entity=self.entity, offset=0, length=1)
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}),
            format='json',
            data={'text': 'One day, the king told him to go save a princess'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Transcriptions with entities cannot be modified.'
        })

    def test_put(self):
        self.client.force_login(self.user)
        manual_tr_id = self.manual_transcription.id
        response = self.client.put(
            reverse('api:transcription-edit', kwargs={'pk': manual_tr_id}),
            format='json',
            data={
                'text': 'a knight was living lonely',
                'orientation': 'vertical-rl',
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.manual_transcription.refresh_from_db()
        self.assertDictEqual(response.json(), {
            'id': str(manual_tr_id),
            'confidence': None,
            'text': 'a knight was living lonely',
            'orientation': 'vertical-rl',
            'worker_version_id': None,
            'worker_run': None,
        })

    def test_put_not_manual(self):
        # self.user is the admin of the corpus, he can edit a worker transcription
        self.client.force_login(self.user)
        response = self.client.put(
            reverse('api:transcription-edit', kwargs={'pk': self.ml_transcription.id}),
            format='json',
            data={'text': 'Once upon a time in a castle', 'orientation': 'horizontal-rl'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data.get('id'), str(self.ml_transcription.id))
        self.assertEqual(data.get('text'), 'Once upon a time in a castle')

    def test_put_internal_user(self):
        self.private_read_user.is_internal = True
        self.private_read_user.save()

        self.client.force_login(self.private_read_user)
        response = self.client.put(
            reverse('api:transcription-edit', kwargs={'pk': self.ml_transcription.id}),
            format='json',
            data={'text': 'He used to play the Bumbulum', 'orientation': 'horizontal-lr'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data.get('id'), str(self.ml_transcription.id))
        self.assertEqual(data.get('text'), 'He used to play the Bumbulum')

    def test_put_invalid_orientation(self):
        """
        An invalid text orientation value causes an error
        """
        self.client.force_login(self.user)
        manual_tr_id = self.manual_transcription.id
        response = self.client.put(
            reverse('api:transcription-edit', kwargs={'pk': manual_tr_id}),
            format='json',
            data={
                'text': 'a knight was living lonely',
                'orientation': 'wobbly',
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'orientation': ['Value is not of type TextOrientation']})

    def test_put_write_right(self):
        """
        A write right is required to put a manual transcription
        """
        self.client.force_login(self.private_read_user)
        response = self.client.put(
            reverse('api:transcription-edit', kwargs={'pk': self.private_transcription.id}),
            format='json',
            data={'text': 'He used to play the Bumbulum', 'orientation': 'horizontal-lr'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'A write access to transcription element corpus is required.'
        })

    def test_put_admin_right(self):
        """
        Updating a transcription produced by a ML worker is forbidden
        """
        self.client.force_login(self.write_user)
        response = self.client.put(
            reverse('api:transcription-edit', kwargs={'pk': self.ml_transcription.id}),
            format='json',
            data={'text': 'He used to play the Bumbulum', 'orientation': 'horizontal-lr'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Only admins can edit non-manual transcription.'
        })

    def test_put_entities(self):
        """
        Updating a transcription that has entities is forbidden
        """
        self.client.force_login(self.user)
        self.manual_transcription.transcription_entities.create(entity=self.entity, offset=0, length=1)
        response = self.client.put(
            reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}),
            format='json',
            data={'text': 'One day, the king told him to go save a princess'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Transcriptions with entities cannot be modified.'
        })

    def test_delete_manual(self):
        """
        Allows to delete a manual transcription with a write right
        """
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(Transcription.objects.filter(id=self.manual_transcription.id).exists())

    def test_delete_manual_entities(self):
        """
        Allows to delete a manual transcription with a write right, even when it has entities
        """
        self.client.force_login(self.user)
        self.manual_transcription.transcription_entities.create(entity=self.entity, offset=0, length=1)
        response = self.client.delete(reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(Transcription.objects.filter(id=self.manual_transcription.id).exists())

    def test_delete_manual_write_right(self):
        """
        Deleting a transcription without a write right is forbidden
        """
        self.client.force_login(self.private_read_user)
        response = self.client.delete(reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'A write access to transcription element corpus is required.'
        })
