import uuid
from datetime import datetime, timedelta, timezone

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element, EntityType, MetaType
from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestParentsElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.surface = cls.corpus.elements.get(name='Surface A')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.private_elt = cls.private_corpus.elements.create(type=cls.private_corpus.types.create(slug='type'))
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def test_element_parents(self):
        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1', ]
        )

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?folder')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1', ]
        )

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?folder=false')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?type=act')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.surface.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 1)
        self.assertEqual(response.json()['results'][0]['name'], 'Act 1')

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.surface.id)}) + '?recursive')
        self.assertEqual(len(response.json()['results']), 2)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Act 1', 'Volume 1'],
        )

        response = self.client.get(
            reverse('api:elements-parents', kwargs={'pk': str(self.surface.id)})
            + '?recursive&name=aCT'
        )
        self.assertEqual(len(response.json()['results']), 1)
        self.assertEqual(response.json()['results'][0]['name'], 'Act 1')

    def test_parents_with_corpus_false(self):
        response = self.client.get(
            reverse('api:elements-parents', kwargs={'pk': str(self.page.id)})
            + '?with_corpus=false'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(response.json()['results'], [
            {
                "id": str(self.vol.id),
                "name": self.vol.name,
                "type": self.vol.type.slug,
                "worker_version_id": None,
                "worker_run": None,
                "confidence": None,
                "zone": None,
                "created": "2020-02-02T01:23:45.678000Z",
                "classes": None,
                "has_children": None,
                "metadata": None,
                "thumbnail_url": self.vol.thumbnail.s3_url,
                "rotation_angle": 0,
                "mirrored": False,
            }
        ])

    def test_parents_worker_version(self):
        """
        List parents filtered by worker_version
        """
        self.corpus.elements.filter(name__contains='Volume 1').update(worker_version=self.worker_version)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'worker_version': str(self.worker_version.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1'],
        )

    def test_parents_worker_version_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'worker_version': 'blah'},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'worker_version': uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})

    def test_parents_with_has_children(self):
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.surface.id)}),
                data={'recursive': True, 'with_has_children': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['has_children'] for element in response.json()['results']},
            {
                'Act 1': True,
                'Volume 1': True,
            }
        )

    def test_parents_modified_since_bad_format(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                HTTP_IF_MODIFIED_SINCE='AAAAAAAAAAAAAAAAA',
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'If-Modified-Since': ['Bad date format']})

    def test_parents_modified_since_not_modified(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_304_NOT_MODIFIED)

    def test_parents_modified_since(self):
        self.corpus.elements.filter(name='Volume 1').update(updated='2099-04-02T13:37:43Z')
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1', ]
        )

    def test_parents_filter_metadata_name(self):
        element = self.corpus.elements.create(type=self.vol.type, name='Element with metadata')
        element.metadatas.create(type=MetaType.Text, name='folio', value='42')
        self.page.add_parent(element)

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'metadata_name': 'folio'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Element with metadata']
        )

    def test_parents_filter_metadata_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'metadata_value': '1v'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['This filter is not supported without metadata_name.'],
        })

    def test_parents_filter_metadata_name_value(self):
        element = self.corpus.elements.create(type=self.vol.type, name='Element with metadata')
        element.metadatas.create(type=MetaType.Text, name='folio', value='42')
        self.vol.metadatas.create(type=MetaType.Text, name='folio', value='43')
        self.page.add_parent(element)

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'metadata_name': 'folio', 'metadata_value': '42'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Element with metadata']
        )

    def test_parents_filter_metadata_operator(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'metadata_operator': 'gt'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_operator': ['This option is not supported without metadata_value.']
        })

    def test_parents_filter_metadata_operator_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'metadata_operator': 'gt', 'metadata_value': 4}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['This filter is not supported without metadata_name.'],
        })

    def test_parents_filter_metadata_operator_invalid(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'metadata_operator': 'aaaaaa'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_operator': [
                'This metadata operator is not supported.',
                'This option is not supported without metadata_value.',
            ]
        })

    def test_parents_filter_numeric_metadata_invalid_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'metadata_name': 'a', 'metadata_value': 'lol', 'metadata_operator': 'gt'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['Value must be a float when using a numeric comparison operator.'],
        })

    def test_parents_filter_numeric_metadata(self):
        element_type = self.corpus.types.get(slug='volume')
        for i in range(1, 6):
            element = self.corpus.elements.create(type=element_type, name=f'Element {i}')
            self.page.add_parent(element)
            element.metadatas.create(name='number', type=MetaType.Numeric, value=str(i))
            element.metadatas.create(name='number', type=MetaType.Text, value='a')

        # operator, value, expected element names
        parameters = [
            ('lt', 2, ['Element 1']),
            ('lte', 2, ['Element 1', 'Element 2']),
            ('gt', 3, ['Element 4', 'Element 5']),
            ('gte', 3, ['Element 3', 'Element 4', 'Element 5']),
        ]

        for operator, value, expected_elements in parameters:
            with self.subTest(operator=operator, value=value):
                with self.assertNumQueries(4):
                    response = self.client.get(
                        reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                        data={
                            'metadata_name': 'number',
                            'metadata_value': value,
                            'metadata_operator': operator,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected_elements
                )

    def test_parents_filter_metadata_contains(self):
        element = self.corpus.elements.create(type=self.vol.type, name='Element with metadata')
        element.metadatas.create(type=MetaType.Text, name='recipe', value='Fry the element')
        self.vol.metadatas.create(type=MetaType.Text, name='recipe', value='Bake the element at 451°F')
        self.page.add_parent(element)

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={
                    'metadata_name': 'recipe',
                    'metadata_operator': 'contains',
                    'metadata_value': 'fry',
                }
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Element with metadata']
        )

    def test_parents_filter_rotation_angle(self):
        element_type = self.corpus.types.first()
        not_rotated = self.corpus.elements.create(type=element_type, name='not rotated')
        rotated = self.corpus.elements.create(type=element_type, name='rotated', rotation_angle=180)
        self.page.add_parent(not_rotated)
        self.page.add_parent(rotated)

        cases = [
            (0, ['not rotated']),
            (1, []),
            (180, ['rotated']),
            (181, [])
        ]
        for rotation_angle, expected_elements in cases:
            with self.subTest(rotation_angle=rotation_angle):
                # Requests that return no elements only make 2 SQL queries
                num_queries = 4 if len(expected_elements) else 2
                with self.assertNumQueries(num_queries):
                    response = self.client.get(
                        reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                        data={
                            'name': 'rotated',
                            'rotation_angle': rotation_angle,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected_elements
                )

    def test_parents_filter_rotation_angle_invalid(self):
        cases = [
            ('to the left', "invalid literal for int() with base 10: 'to the left'"),
            (-1, 'A rotation angle must be between 0 and 359 degrees'),
            (360, 'A rotation angle must be between 0 and 359 degrees'),
            (54.2, "invalid literal for int() with base 10: '54.2'"),
            (4.14j, "invalid literal for int() with base 10: '4.14j'"),
        ]
        for rotation_angle, expected_error in cases:
            with self.subTest(rotation_angle=rotation_angle):
                with self.assertNumQueries(1):
                    response = self.client.get(
                        reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                        data={
                            'rotation_angle': rotation_angle,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {
                    'rotation_angle': [expected_error]
                })

    def test_parents_filter_mirrored(self):
        element_type = self.corpus.types.first()
        mirrored = self.corpus.elements.create(type=element_type, name='mirrored', mirrored=True)
        not_mirrored = self.corpus.elements.create(type=element_type, name='not mirrored', mirrored=False)
        self.page.add_parent(mirrored)
        self.page.add_parent(not_mirrored)

        cases = [
            (False, ['not mirrored']),
            (True, ['mirrored']),
        ]
        for mirrored, expected_elements in cases:
            with self.subTest(mirrored=mirrored):
                with self.assertNumQueries(4):
                    response = self.client.get(
                        reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                        data={
                            'name': 'mirrored',
                            'mirrored': mirrored,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected_elements
                )

    def test_parents_invalid_order(self):
        cases = [
            ({'order': 'blah', 'order_direction': 'asc'}, {'order': ['Unknown sorting field']}),
            ({'order': 'name', 'order_direction': 'left'}, {'order_direction': ['Unknown sorting direction']}),
            ({'order': 'blah', 'order_direction': 'left'}, {'order': ['Unknown sorting field']}),
        ]
        for params, expected in cases:
            with self.subTest(**params):
                response = self.client.get(
                    reverse('api:elements-parents', kwargs={'pk': self.page.id}),
                    params,
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected)

    def test_parents_order_created(self):
        elements = list(Element.objects.get_ascending(self.page.id).order_by('?').only('id'))
        created = datetime(2020, 2, 2, tzinfo=timezone.utc)
        for element in elements:
            element.created = created
            created += timedelta(seconds=1)
        Element.objects.bulk_update(elements, ['created'])

        cases = [
            ('asc', elements[:20]),
            ('desc', reversed(elements[-20:])),
        ]
        for direction, expected in cases:
            with self.subTest(direction=direction), self.assertNumQueries(4):
                response = self.client.get(
                    reverse('api:elements-parents', kwargs={'pk': self.page.id}),
                    {'order': 'created', 'order_direction': direction, 'recursive': True}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['id'] for element in response.json()['results']],
                    [str(element.id) for element in expected]
                )

    def test_parents_order_random(self):
        # Create more parent elements to make a random order more visible
        for i in range(10):
            element = self.corpus.elements.create(type=self.vol.type, name=f'Element {i}')
            self.page.add_parent(element)

        # Get all expected element IDs as strings
        element_ids = list(map(str, Element.objects.get_ascending(self.page.id).values_list('id', flat=True)))

        # Repeat the API request multiple times; the element order should change at least once
        responses = set()
        for _ in range(5):
            with self.assertNumQueries(4):
                response = self.client.get(
                    reverse('api:elements-parents', kwargs={'pk': self.page.id}),
                    {'order': 'random', 'recursive': True},
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)

            data = response.json()

            # Checking for the expected element IDs will start failing randomly
            # if there is more than 1 page of results.
            self.assertIsNone(data['next'])
            actual_ids = [element['id'] for element in data['results']]
            self.assertCountEqual(element_ids, actual_ids)

            responses.add(tuple(actual_ids))

        # We know that all responses have the expected elements, so if the responses set
        # has more than one item, the ordering has changed between requests, so it is likely to be random.
        self.assertGreater(len(responses), 1)

    def test_parents_with_metadata(self):
        act_meta = Element.objects.get(name='Act 1').metadatas.get()
        vol_meta = self.vol.metadatas.create(
            name='Never gonna give you up',
            type=MetaType.Text,
            value='Never gonna let you down',
        )
        person_type = EntityType.objects.get(name='person', corpus=self.corpus)
        # Create a metadata with an entity, which should be excluded
        self.vol.metadatas.create(
            name='Never gonna run around',
            type=MetaType.Text,
            value='and desert you',
            entity=self.corpus.entities.create(name='Rick Astley', type=person_type)
        )

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.surface.id)}),
                {'with_metadata': 'true', 'recursive': 'true'},
            )
        self.assertDictEqual({
            element['name']: element['metadata']
            for element in response.json()['results']
        }, {
            'Act 1': [
                {
                    'id': str(act_meta.id),
                    'name': 'number',
                    'type': 'text',
                    'value': '1',
                    'dates': [],
                }
            ],
            'Volume 1': [
                {
                    'id': str(vol_meta.id),
                    'name': 'Never gonna give you up',
                    'type': 'text',
                    'value': 'Never gonna let you down',
                    'dates': [],
                },
            ]
        })

    def test_transcription_confidence_invalid(self):
        cases = [
            ({'transcription_confidence': 'very'}, {'transcription_confidence': ["could not convert string to float: 'very'"]}),
            ({'transcription_confidence': 'nan'}, {'transcription_confidence': ['Confidence must be between 0 and 1']}),
            ({'transcription_confidence': 'inf'}, {'transcription_confidence': ['Confidence must be between 0 and 1']}),
            ({'transcription_confidence': '42'}, {'transcription_confidence': ['Confidence must be between 0 and 1']}),
            ({'transcription_confidence': '-.5'}, {'transcription_confidence': ['Confidence must be between 0 and 1']}),
            (
                {'transcription_confidence_operator': 'hah'},
                {'transcription_confidence_operator': [
                    'This operator is not supported.',
                    'This option is not supported without transcription_confidence.',
                ]}
            ),
            (
                {'transcription_confidence': '0.3', 'transcription_confidence_operator': 'lol'},
                {'transcription_confidence_operator': ['This operator is not supported.']}
            )
        ]
        for data, expected_errors in cases:
            with self.subTest(**data):
                response = self.client.get(reverse('api:elements-parents', kwargs={'pk': self.page.id}), data)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)

    def test_transcription_confidence_filter(self):
        act = Element.objects.get(name='Act 1')
        act.transcriptions.create(text='blah', confidence=0.8)
        self.vol.transcriptions.create(text='bleh', confidence=0.4)

        cases = [
            (
                {'transcription_confidence': 0.4},
                ['Volume 1']
            ),
            (
                {'transcription_confidence': 0.5},
                []
            ),
            (
                {'transcription_confidence': 0.4, 'transcription_confidence_operator': 'gt'},
                ['Act 1']
            ),
            (
                {'transcription_confidence': 0.4, 'transcription_confidence_operator': 'lte'},
                ['Volume 1']
            ),
            (
                {'transcription_confidence': 0.4, 'transcription_confidence_operator': 'gte'},
                ['Act 1', 'Volume 1']
            )
        ]
        for data, element_names in cases:
            with self.subTest(**data):
                response = self.client.get(
                    reverse('api:elements-parents', kwargs={'pk': self.surface.id}),
                    {'recursive': True, **data}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    element_names,
                )

    def test_transcription_worker_version_invalid(self):
        cases = [
            ('a', {'transcription_worker_version': ['Invalid UUID']}),
            (42, {'transcription_worker_version': ['Invalid UUID']}),
            (True, {'transcription_worker_version': ['Invalid UUID']}),
            ('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa', {'transcription_worker_version': ['This worker version does not exist.']}),
        ]
        for transcription_worker_version, expected_errors in cases:
            with self.subTest(transcription_worker_version=transcription_worker_version):
                response = self.client.get(
                    reverse('api:elements-parents', kwargs={'pk': self.page.id}),
                    {'transcription_worker_version': transcription_worker_version}
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)

    def test_transcription_worker_version_filter(self):
        act = Element.objects.get(name='Act 1')
        other_act = self.corpus.elements.create(name='Other act', type=act.type)
        self.surface.add_parent(other_act)

        dla = WorkerVersion.objects.get(worker__slug='dla')
        act.transcriptions.create(text='blah', confidence=1, worker_version=self.worker_version)
        other_act.transcriptions.create(text='bloh', confidence=1, worker_version=dla)
        self.vol.transcriptions.create(text='bleh', confidence=1, worker_version=None)

        cases = [
            (str(self.worker_version.id), ['Act 1']),
            (str(dla.id), ['Other act']),
            (False, ['Volume 1']),
        ]
        for transcription_worker_version, element_names in cases:
            with self.subTest(transcription_worker_version=transcription_worker_version):
                response = self.client.get(
                    reverse('api:elements-parents', kwargs={'pk': self.surface.id}),
                    {'transcription_worker_version': transcription_worker_version, 'recursive': True}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    element_names,
                )

    def test_confidence_invalid(self):
        cases = [
            ({'confidence': 'very'}, {'confidence': ["could not convert string to float: 'very'"]}),
            ({'confidence': 'nan'}, {'confidence': ['Confidence must be between 0 and 1']}),
            ({'confidence': 'inf'}, {'confidence': ['Confidence must be between 0 and 1']}),
            ({'confidence': '42'}, {'confidence': ['Confidence must be between 0 and 1']}),
            ({'confidence': '-.5'}, {'confidence': ['Confidence must be between 0 and 1']}),
            (
                {'confidence_operator': 'hah'},
                {'confidence_operator': [
                    'This operator is not supported.',
                    'This option is not supported without `confidence`.',
                ]}
            ),
            (
                {'confidence': '0.3', 'confidence_operator': 'lol'},
                {'confidence_operator': ['This operator is not supported.']}
            )
        ]
        for data, expected_errors in cases:
            with self.subTest(**data):
                response = self.client.get(reverse('api:elements-parents', kwargs={'pk': self.page.id}), data)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)

    def test_confidence_filter(self):
        elements = Element.objects.bulk_create([
            Element(
                corpus=self.corpus,
                type=self.corpus.types.get(slug='volume'),
                confidence=i / 10,
                name=f'Confidence {i / 10}'
            )
            for i in range(11)
        ])
        for element in elements:
            self.page.add_parent(element)

        cases = [
            (
                {'confidence': 0.5},
                ['Confidence 0.5']
            ),
            (
                {'confidence': 0.51},
                []
            ),
            (
                {'confidence': 0.5, 'confidence_operator': 'gt'},
                ['Confidence 0.6', 'Confidence 0.7', 'Confidence 0.8', 'Confidence 0.9', 'Confidence 1.0']
            ),
            (
                {'confidence': 0.5, 'confidence_operator': 'lte'},
                ['Confidence 0.0', 'Confidence 0.1', 'Confidence 0.2', 'Confidence 0.3', 'Confidence 0.4', 'Confidence 0.5']
            )
        ]
        for data, element_names in cases:
            with self.subTest(**data):
                response = self.client.get(reverse('api:elements-parents', kwargs={'pk': self.page.id}), data)
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    element_names,
                )
