import re
from collections import namedtuple
from uuid import UUID

import yaml
from django.core.validators import URLValidator

TaskRecipe = namedtuple(
    "TaskRecipe",
    "image, parents, command, shm_size, environment, tags, has_docker_socket, artifact, requires_gpu, extra_files",
)


def validate_task(task, top_env):
    """
    Validate a task description
    - image (required)
    - command (optional)
    - parents (optional)
    - shm_size (optional)
    - env (optional)
    - tags (optional)
    - has_docker_socket (optional)
    - artifact (optional)
    - requires_gpu (optional)
    - extra_files (optional)
    """
    assert isinstance(task, dict), "Task should be a dict"
    assert "image" in task, "Missing image"

    if "parents" in task:
        assert isinstance(task["parents"], list), "Task parents should be a list"
    if "tags" in task:
        assert isinstance(task["tags"], list), "Task tags should be a list"
    if task.get("shm_size"):
        value = str(task["shm_size"])
        assert re.match(
            r"^[0-9]+[bkmgBKMG]?$", str(value)
        ), f"{str(value)} is not a valid value for shm_size"
        if str(value).isdigit():
            assert int(value) > 0, "shm_size value must be greater than 0"
        else:
            assert int(value[:-1]) > 0, "shm_size value must be greater than 0"

    # Add optional local environment variables
    env = top_env.copy()
    if "env" in task:
        assert isinstance(
            task["env"], dict
        ), "Task environment variables should be a dict"
        env.update(task["env"])

    if "has_docker_socket" in task:
        assert isinstance(
            task["has_docker_socket"], bool
        ), "Task has_docker_socket should be a boolean"

    if "artifact" in task:
        assert isinstance(task["artifact"], str), "Task artifact should be a string"
        try:
            UUID(task["artifact"])
        except (TypeError, ValueError):
            raise AssertionError("Task artifact should be a valid UUID string")

    if "requires_gpu" in task:
        assert isinstance(
            task["requires_gpu"], bool
        ), "Task requires_gpu should be a boolean"

    if "extra_files" in task:
        assert isinstance(
            task["extra_files"], dict
        ), "Task extra_files should be a dict of strings/url"
        url_validator = URLValidator(schemes=["http", "https"])
        for key, value in task["extra_files"].items():
            assert isinstance(key, str), "All Task extra_files keys should be strings"
            assert isinstance(
                value, str
            ), "All Task extra_files values should be strings"
            url_validator(value)

    return TaskRecipe(
        task["image"],
        task.get("parents", []),
        task.get("command"),
        task.get("shm_size", None),
        env,
        task.get("tags", []),
        task.get("has_docker_socket", False),
        task.get("artifact", ""),
        task.get("requires_gpu", False),
        task.get("extra_files", {}),
    )


def parse_recipe(recipe):
    """
    Parse a recipe and check its content
    Build a dict with all recipes per slug
    """
    content = yaml.safe_load(recipe)
    assert isinstance(content, dict), "Recipe should be a dict"

    # Load optional environment variables
    env = content.get("env", {})
    assert isinstance(env, dict)
    assert all(map(lambda x: isinstance(x, str), env.keys()))
    env = {k: str(v) for k, v in env.items()}

    # Load tasks
    tasks = content.get("tasks")
    assert tasks, "No tasks"
    assert isinstance(tasks, dict), "Tasks should be a dict"
    assert all(tasks), "Tasks should have non-blank slugs"

    # Validate all tasks
    tasks = {slug: validate_task(task, env) for slug, task in tasks.items()}
    assert len(tasks) > 0, "No tasks recipes"

    # Check all parents exists
    for slug, task in tasks.items():
        for parent in task.parents:
            assert parent in tasks, "Missing parent {} for task {}".format(parent, slug)

    return env, tasks


def recipe_depth(slug, recipes, depth=0):
    """
    Find recursively the depth of a recipe
    """
    recipe = recipes.get(slug)
    assert recipe is not None, "Recipe not found"
    assert isinstance(recipe, TaskRecipe), "Recipe is not a TaskRecipe"

    if not recipe.parents:
        return depth

    return max((recipe_depth(parent, recipes, depth + 1) for parent in recipe.parents))
