from unittest.mock import patch

from django.test import TestCase
from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from arkindex.ponos.models import Agent, Farm, State, Workflow

ONE_TASK = """
tasks:
  test:
    image: hello-world
"""

THREE_TASKS = """
tasks:
  test:
    image: hello-world
  test-subsub:
    image: XXX
    parents:
    - test-sub
  test-sub:
    image: YYY:latest
    parents:
    - test
"""

MULTIPLE_PARENTS_TASK = """
tasks:
  test:
    image: hello-world
    parents:
    - test_aaa
    - test_bbb
  test_aaa:
    image: a
  test_bbb:
    image: b
"""

TASK_SHM_SIZE = """
tasks:
  test:
    image: hello-world
    shm_size: 505m
"""

TASK_INVALID_SHM_SIZE = """
tasks:
  test:
    image: hello-world
    shm_size: douze
"""

TASK_SHM_SIZE_NONE = """
tasks:
  test:
    image: hello-world
    shm_size: null
"""


class WorkflowTestCase(TestCase):
    """
    Create some workflows & tasks
    """

    @classmethod
    def setUpTestData(cls):
        cls.farm = Farm.objects.create(name="testfarm")
        # Create a fake Agent instance
        cls.agent = Agent.objects.create(
            hostname="test_agent",
            cpu_cores=2,
            cpu_frequency=1e9,
            public_key="",
            farm=cls.farm,
            ram_total=2e9,
            last_ping=timezone.now(),
        )

    def test_workflow_start_one_task(self):
        w = Workflow.objects.create(farm=self.farm, recipe=ONE_TASK)
        tasks = w.tasks
        self.assertEqual(tasks.count(), 0)

        w.start()
        self.assertEqual(tasks.count(), 1)
        task = tasks.first()
        self.assertFalse(task.parents.exists())
        self.assertEqual(task.run, 0)
        self.assertEqual(task.depth, 0)

    def test_workflow_start_three_tasks(self):
        w = Workflow.objects.create(farm=self.farm, recipe=THREE_TASKS)
        tasks = w.tasks
        self.assertEqual(tasks.count(), 0)

        w.start()
        self.assertEqual(tasks.count(), 3)
        t1, t2, t3 = w.tasks.all()

        self.assertFalse(t1.parents.exists())
        self.assertEqual(t1.run, 0)
        self.assertEqual(t1.depth, 0)

        self.assertEqual(t2.parents.get(), t1)
        self.assertEqual(t2.run, 0)
        self.assertEqual(t2.depth, 1)

        self.assertEqual(t3.parents.get(), t2)
        self.assertEqual(t3.run, 0)
        self.assertEqual(t3.depth, 2)

    def test_multiple_parents_task(self):
        w = Workflow.objects.create(farm=self.farm, recipe=MULTIPLE_PARENTS_TASK)

        w.start()
        self.assertEqual(w.tasks.count(), 3)

        task = w.tasks.get(slug="test")
        self.assertEqual([t.slug for t in task.parents.all()], ["test_aaa", "test_bbb"])

    def test_workflow_running_override(self):
        """
        Test that a single running task in a workflow will override any other state
        """
        w = Workflow.objects.create(farm=self.farm, recipe=THREE_TASKS)
        w.start()

        self.assertEqual(w.tasks.count(), 3)
        self.assertEqual(w.state, State.Unscheduled)
        t1, t2, t3 = w.tasks.all()

        t1.state = State.Running
        t1.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Error
        t2.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Failed
        t2.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Stopping
        t2.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Stopped
        t2.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Pending
        t2.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Completed
        t3.state = State.Completed
        t2.save()
        t3.save()
        self.assertEqual(w.state, State.Running)

    @patch("arkindex.ponos.models.s3")
    @patch("arkindex.ponos.models.Task.s3_logs_get_url")
    def test_task_parents_update(self, s3_mock, s3_logs_mock):
        w = Workflow.objects.create(farm=self.farm, recipe=MULTIPLE_PARENTS_TASK)
        w.start()

        child = w.tasks.get(slug="test")
        parent1, parent2 = w.tasks.exclude(slug="test")

        response = self.client.get(
            reverse("api:agent-actions"),
            HTTP_AUTHORIZATION="Bearer {}".format(self.agent.token.access_token),
            data={"cpu_load": 0.99, "ram_load": 0.49},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        actions = response.json()["actions"]
        self.assertEqual(len(actions), 1)
        self.assertNotEqual(actions[0]["task_id"], str(child.id))

        parent1.state = State.Completed
        parent1.save()

        parent2.agent = self.agent
        parent2.save()

        self.client.get(
            reverse("api:agent-actions"),
            HTTP_AUTHORIZATION="Bearer {}".format(self.agent.token.access_token),
        )
        # Completing 2nd parent should change test_task state to pending
        response = self.client.patch(
            reverse("api:task-details", kwargs={"pk": str(parent2.id)}),
            HTTP_AUTHORIZATION="Bearer {}".format(self.agent.token.access_token),
            data={"state": State.Completed.value},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        child.refresh_from_db()
        self.assertEqual(child.state, State.Pending)
        self.assertEqual(w.state, State.Pending)

    def test_workflow_retry_updates_finished(self):
        w = Workflow.objects.create(farm=self.farm, recipe=ONE_TASK)
        w.start()
        w.tasks.update(state=State.Completed)
        w.finished = timezone.now()
        w.save()

        with self.assertNumQueries(7):
            w.retry()
        self.assertEqual(w.tasks.count(), 2)
        self.assertIsNone(w.finished)

    def test_workflow_stop_unscheduled_updates_finished(self):
        w = Workflow.objects.create(farm=self.farm, recipe=ONE_TASK)
        w.start()
        # All tasks should be now unscheduled.
        self.assertFalse(w.tasks.exclude(state=State.Unscheduled).exists())

        # Stopping the workflow causes all tasks to immediately be Stopped
        w.stop()
        self.assertFalse(w.tasks.exclude(state=State.Stopped).exists())
        self.assertIsNotNone(w.finished)

    def test_workflow_start_shm_size(self):
        w = Workflow.objects.create(farm=self.farm, recipe=TASK_SHM_SIZE)
        tasks = w.tasks
        self.assertEqual(tasks.count(), 0)

        w.start()
        self.assertEqual(tasks.count(), 1)
        task = tasks.first()
        self.assertFalse(task.parents.exists())
        self.assertEqual(task.shm_size, "505m")
        self.assertEqual(task.run, 0)
        self.assertEqual(task.depth, 0)

    def test_workflow_start_invalid_shm_size(self):
        w = Workflow.objects.create(farm=self.farm, recipe=TASK_INVALID_SHM_SIZE)
        tasks = w.tasks
        self.assertEqual(tasks.count(), 0)
        with self.assertRaises(AssertionError) as e:
            w.start()
        self.assertEqual(str(e.exception), "douze is not a valid value for shm_size")

    def test_workflow_start_none_shm_size(self):
        w = Workflow.objects.create(farm=self.farm, recipe=TASK_SHM_SIZE_NONE)
        tasks = w.tasks
        self.assertEqual(tasks.count(), 0)
        w.start()
        self.assertEqual(tasks.count(), 1)
        task = tasks.first()
        self.assertFalse(task.parents.exists())
        self.assertEqual(task.shm_size, None)
        self.assertEqual(task.run, 0)
        self.assertEqual(task.depth, 0)
