from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.process.models import WorkerConfiguration, WorkerRun
from arkindex.process.serializers.imports import ProcessTrainingSerializer
from arkindex.process.serializers.workers import WorkerConfigurationSerializer, WorkerVersionSerializer
from arkindex.training.models import ModelVersion, ModelVersionState
from arkindex.training.serializers import ModelVersionLightSerializer
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level

# To prevent each element worker to retrieve contextual information
# (process, worker version, model version…) with extra GET requests, we
# do serialize all the related information on WorkerRun serializers.


class WorkerRunSerializer(serializers.ModelSerializer):
    """
    Serialize a worker run for creation.
    Worker version and parents are required, configuration is optional.
    """
    parents = serializers.ListField(child=serializers.UUIDField())

    worker_version_id = serializers.UUIDField(write_only=True, source='version_id')
    worker_version = WorkerVersionSerializer(read_only=True, source='version')

    model_version = ModelVersionLightSerializer(read_only=True)

    configuration = WorkerConfigurationSerializer(read_only=True)
    process = ProcessTrainingSerializer(read_only=True)
    configuration_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerConfiguration.objects.all(),
        required=False,
        allow_null=True,
        write_only=True,
        style={'base_template': 'input.html'},
    )

    class Meta:
        model = WorkerRun
        fields = (
            'id',
            'parents',
            'worker_version_id',
            'worker_version',
            'process',
            'configuration_id',
            'configuration',
            'model_version',
        )


class WorkerRunEditSerializer(WorkerRunSerializer):
    """
    Serialize a worker run for edition.
    Process and worker version cannot be edited.
    Parents, model version and configuration can be edited.
    """
    model_version_id = serializers.PrimaryKeyRelatedField(
        queryset=ModelVersion.objects.all().select_related('model'),
        required=False,
        allow_null=True,
        write_only=True,
        source='model_version',
        style={'base_template': 'input.html'},
    )

    class Meta:
        model = WorkerRun
        fields = (
            'id',
            'parents',
            'worker_version',
            'process',
            'configuration_id',
            'configuration',
            'model_version_id',
            'model_version',
        )

    def validate_model_version_id(self, model_version):
        model_usage = self.context.get('model_usage')
        if not model_usage:
            raise ValidationError("This worker version does not support model usage.")

        if model_version.state != ModelVersionState.Available:
            raise ValidationError("This model version is not in an available state.")

        # Check access rights on model version
        access_level = get_max_level(self.context["request"].user, model_version.model)
        if not access_level or access_level < Role.Contributor.value:
            raise ValidationError('You do not have access to this model version.')

        return model_version

    def validate(self, data):
        # Store configuration if provided
        if 'configuration_id' in data:
            data['configuration'] = data['configuration_id']
        return data
