
import importlib
import uuid
from unittest.mock import MagicMock, call, patch

import yaml
from django.test import override_settings
from django.urls import reverse
from rest_framework import status
from rest_framework.exceptions import ValidationError

from arkindex.documents.models import Corpus, ElementType
from arkindex.ponos.models import Farm, State, Task, Workflow
from arkindex.process.models import (
    ActivityState,
    Process,
    ProcessMode,
    WorkerActivity,
    WorkerActivityState,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.process.utils import get_default_farm_id
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.users.models import Role, User

RECIPE = '''
tasks:
  first:
    image: alpine
  second:
    parents:
      - first
    image: hello-world
'''


class TestProcesses(FixtureAPITestCase):
    """
    Test process management
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(url='http://my_repo.fake/workers/worker')
        cls.rev = cls.repo.revisions.get()
        cls.img_df = cls.corpus.files.create(
            name='test.jpg',
            size=42,
            content_type='image/jpg',
        )
        cls.pdf_df = cls.corpus.files.create(
            name='test.pdf',
            size=42,
            content_type='application/pdf',
        )
        cls.iiif_df = cls.corpus.files.create(
            name='test.json',
            size=42,
            content_type='application/json',
        )
        cls.page_type = ElementType.objects.get(corpus=cls.corpus, slug='page')
        cls.volume_type = ElementType.objects.get(corpus=cls.corpus, slug='volume')
        cls.recognizer = WorkerVersion.objects.get(worker__slug='reco')
        cls.dla = WorkerVersion.objects.get(worker__slug='dla')
        cls.version_gpu = WorkerVersion.objects.get(worker__slug='worker-gpu')
        cls.workers_process = cls.corpus.processes.get(mode=ProcessMode.Workers)
        cls.version_with_model = WorkerVersion.objects.get(worker__slug='generic')

        # Create a model and a model version
        cls.model_1 = Model.objects.create(name='My model')
        cls.model_version_1 = ModelVersion.objects.create(model=cls.model_1, state=ModelVersionState.Available, size=8, hash='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa', archive_hash='bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb')

        # Create multiple processes the user can access
        cls.user2 = User.objects.create_user('user2@test.test', display_name='Process creator')

        # Guest access (A user owns a process on a corpus they are not a member anymore)
        cls.user_img_process = Process.objects.create(
            mode=ProcessMode.Files,
            creator=cls.user,
            corpus=Corpus.objects.create(name='Private'),
        )
        cls.private_corpus = Corpus.objects.get(name='Private')
        # Contributor access
        cls.repo.memberships.create(user=cls.user, level=Role.Contributor.value)
        cls.repository_process = Process.objects.create(
            mode=ProcessMode.Repository,
            creator=cls.user2,
            revision=cls.rev,
        )
        # Admin access
        cls.elts_process = cls.corpus.processes.create(mode=ProcessMode.Workers, creator=cls.user2)
        cls.processes = Process.objects.filter(
            id__in=[cls.user_img_process.id, cls.repository_process.id, cls.elts_process.id]
        )
        # Training process
        cls.train, cls.val, cls.test = (
            cls.corpus.elements.create(name=name, type=cls.volume_type)
            for name in ('train', 'val', 'test')
        )
        cls.training_model = Model.objects.create(name='Generic model', public=False)
        cls.training_process = Process.objects.create(
            creator=cls.user,
            corpus=cls.corpus,
            mode=ProcessMode.Training,
            workflow=cls.elts_process.workflow,
            model=cls.training_model,
            train_folder=cls.train,
            validation_folder=cls.val,
            test_folder=cls.test,
        )

        cls.local_process = Process.objects.create(
            creator=cls.user,
            mode=ProcessMode.Local,
        )
        cls.local_process_2 = Process.objects.create(
            creator=cls.user2,
            mode=ProcessMode.Local,
        )

        cls.import_worker_version = WorkerVersion.objects.get(worker__slug='file_import')

    def _serialize_process(self, process):
        updated = process.updated
        if process.workflow:
            updated = max(updated, process.workflow.tasks.order_by('-updated').first().updated)
        return {
            'name': process.name,
            'id': str(process.id),
            'state': process.state.value,
            'mode': process.mode.value,
            'corpus': process.corpus_id and str(process.corpus.id),
            'workflow': process.workflow and f'http://testserver/api/v1/workflow/{process.workflow.id}/',
            'activity_state': process.activity_state.value,
            'created': process.created.isoformat().replace('+00:00', 'Z'),
            'updated': updated.isoformat().replace('+00:00', 'Z'),
            'finished': (
                process.workflow.finished.isoformat().replace('+00:00', 'Z')
                if process.workflow and process.workflow.finished
                else None
            ),
        }

    def build_task(self, workflow_id, run, state, depth=1):
        """
        Create a simple task with its workflow and process if required
        """
        workflow, created = Workflow.objects.get_or_create(
            id=workflow_id,
            farm=Farm.objects.get(name="Wheat farm"),
            defaults={'recipe': RECIPE}
        )
        if created:
            # Creates the associated process
            Process.objects.create(
                corpus=self.corpus,
                workflow=workflow,
                mode=ProcessMode.Workers,
                creator=self.user,
            )
        return Task.objects.create(run=run, depth=depth, workflow=workflow, state=state, slug=str(uuid.uuid4()))

    def test_list_requires_login(self):
        response = self.client.get(reverse('api:process-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        """
        A user can list processes they have access to
        """
        # Order processes by causing updates on them
        self.elts_process.start()
        self.user_img_process.save()
        self.repository_process.save()
        # An update on a process task will make it first
        task = self.elts_process.workflow.tasks.first()
        task.save()

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:process-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # Processes are ordered
        self.assertDictEqual(response.json(), {
            'count': 5,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                self._serialize_process(p)
                for p in [
                    self.elts_process,
                    self.repository_process,
                    self.user_img_process,
                    self.training_process,
                    self.workers_process,
                ]
            ]
        })

    def test_list_with_workflow(self):
        """
        Filter processes that have a workflow i.e. that have been started
        """
        self.user_img_process.start()
        self.client.force_login(self.user)
        self.user_img_process.activity_state = ActivityState.Ready
        self.user_img_process.save()
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:process-list'), {'with_workflow': 'true'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        results = data['results']
        self.assertListEqual(results, [self._serialize_process(self.user_img_process)])

    def test_list_exclude_workflow(self):
        """
        Filter process that have not been started yet
        """
        self.repository_process.start()
        self.elts_process.start()
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-list'), {'with_workflow': 'false'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 3,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                self._serialize_process(self.training_process),
                self._serialize_process(self.user_img_process),
                self._serialize_process(self.workers_process),
            ]
        })

    def test_list_filter_corpus(self):
        self.client.force_login(self.superuser)
        corpus2 = Corpus.objects.create(name='Another corpus', description='something')
        process2 = corpus2.processes.create(creator=self.user, mode=ProcessMode.Files)

        response = self.client.get(reverse('api:process-list'), {'corpus': str(corpus2.id), 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(process2.id))

    def test_list_filter_corpus_invalid(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:process-list'), {'corpus': 'oh-no'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'corpus': ["'oh-no' is not a valid UUID"]})

    def test_list_filter_mode(self):
        self.client.force_login(self.user)
        assert Process.objects.exclude(mode=ProcessMode.Files).exists()
        response = self.client.get(reverse('api:process-list'), {'mode': ProcessMode.Files.value, 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(self.user_img_process.id))

    def test_list_filter_mode_local(self):
        """
        Filtering by the local mode shows only the local process of the current user
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-list'), {'mode': ProcessMode.Local.value})
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                self._serialize_process(self.local_process),
            ]
        })

    def test_list_filter_unexisting_mode(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:process-list'), {'mode': 'unexisting_mode'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'mode': ["Mode 'unexisting_mode' does not exist"]})

    def test_list_filter_created(self):
        """
        Display processes that have been created by the user
        """
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:process-list'), {'created': 'true'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 3,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                self._serialize_process(self.training_process),
                self._serialize_process(self.user_img_process),
                self._serialize_process(self.workers_process),
            ]
        })

    def test_list_filter_not_created(self):
        """
        Display processes that have not been created by the user
        """
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:process-list'), {'created': 'false'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 2)
        self.assertCountEqual(
            [p['id'] for p in data['results']],
            [str(p.id) for p in self.processes.exclude(creator=self.user)]
        )

    def test_list_filter_id(self):
        self.client.force_login(self.user)

        # Generate a UUID whose first characters are not the same as the first Process
        process_id = str(uuid.uuid4())
        while process_id[:10] == str(self.user_img_process.id)[:10]:
            process_id = str(uuid.uuid4())

        process2 = self.corpus.processes.create(
            id=process_id,
            creator=self.user,
            mode=ProcessMode.Files,
        )
        response = self.client.get(reverse('api:process-list'), {'id': process_id[:10], 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(process2.id))

    def test_list_filter_name(self):
        self.client.force_login(self.user)

        process_id_1 = str(uuid.uuid4())
        process1 = self.corpus.processes.create(
            id=process_id_1,
            creator=self.user,
            mode=ProcessMode.Workers,
            name='Number One'
        )

        process_id_2 = str(uuid.uuid4())
        self.corpus.processes.create(
            id=process_id_2,
            creator=self.user,
            mode=ProcessMode.Workers,
            name='Numero Duo'
        )

        response = self.client.get(reverse('api:process-list'), {'name': 'Numb', 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(process1.id))

    def test_list_no_tasks(self):
        """
        Ensure the Process reports an Unscheduled state when there are no tasks in its workflow
        """
        self.assertIsNone(self.user_img_process.workflow)
        self.user_img_process.start()
        self.user_img_process.workflow.tasks.all().delete()

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:process-list'), {'id': str(self.user_img_process.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['state'], State.Unscheduled.value)

    def test_list_process_filter_wrong_state_value(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:process-list'), {'state': 'spain'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'state': ["State 'spain' does not exist"]})

    def test_list_process_state_filter(self):
        """
        Only last run should be used to filter processes by state
        """
        workflows = [uuid.uuid4() for i in range(3)]
        for task in (
            (workflows[0], 0, State.Failed),
            (workflows[0], 1, State.Completed),

            (workflows[1], 0, State.Error),
            (workflows[1], 0, State.Completed),

            (workflows[2], 0, State.Completed),
            (workflows[2], 1, State.Completed),
            (workflows[2], 2, State.Stopped),
        ):
            self.build_task(*task)
        processes = [Process.objects.get(workflow_id=w) for w in workflows]
        # Only first process is completed
        self.assertListEqual([p for p in processes if p.state.value == 'completed'], [processes[0]])

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:process-list'), {'state': 'completed'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual([r['id'] for r in response.json()['results']], [str(processes[0].id)])

    def test_list_process_state_filter_stopped(self):
        workflows = [uuid.uuid4() for i in range(4)]
        for task in (
            # Should be stopping
            (workflows[0], 0, State.Error),
            (workflows[0], 1, State.Stopping),
            # Should be in error state
            (workflows[1], 0, State.Error),
            (workflows[1], 0, State.Stopping),
            # Should be stopped
            (workflows[2], 0, State.Unscheduled),
            (workflows[2], 0, State.Stopped),
            # Should be stopped
            (workflows[3], 0, State.Failed),
            (workflows[3], 1, State.Pending),
            (workflows[3], 1, State.Stopped),
        ):
            self.build_task(*task)
        processes = [Process.objects.get(workflow_id=w) for w in workflows]
        self.assertListEqual([p for p in processes if p.state.value == 'stopped'], [processes[2], processes[3]])

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:process-list'), {'state': 'stopped'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            [r['id'] for r in response.json()['results']],
            [str(processes[2].id), str(processes[3].id)]
        )

    def test_list_process_state_filter_unscheduled(self):
        """
        Processes with no task should be interpreted as unscheduled
        """
        workflow_id = uuid.uuid4()
        self.build_task(workflow_id, 0, State.Completed)
        self.build_task(workflow_id, 1, State.Unscheduled)
        unscheduled_with_workflow = Process.objects.get(workflow_id=workflow_id)

        # Workflows with no tasks should be considered unscheduled
        self.assertIsNone(self.user_img_process.workflow_id)
        self.assertEqual(self.user_img_process.state, State.Unscheduled)
        self.assertIsNone(self.repository_process.workflow_id)
        self.assertEqual(self.repository_process.state, State.Unscheduled)
        self.assertIsNone(self.elts_process.workflow_id)
        self.assertEqual(self.elts_process.state, State.Unscheduled)

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:process-list'), {'state': 'unscheduled', 'with_workflow': True})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                self._serialize_process(unscheduled_with_workflow)
            ]
        })

        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(reverse('api:process-list'), {'state': 'unscheduled', 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 5,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                self._serialize_process(self.training_process),
                self._serialize_process(self.elts_process),
                self._serialize_process(self.repository_process),
                self._serialize_process(self.user_img_process),
                self._serialize_process(self.workers_process),
            ]
        })

    def test_details_requires_login(self):
        response = self.client.get(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_details_no_tasks(self):
        """
        Ensure the Process reports an Unscheduled state when there are no tasks in its workflow
        """
        self.assertIsNone(self.user_img_process.workflow)
        self.user_img_process.start()
        self.user_img_process.workflow.tasks.all().delete()
        # A user needs access to a process's corpus to get this process's details
        self.private_corpus.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['state'], State.Unscheduled.value)

    def test_details(self):
        self.client.force_login(self.user)
        # A user needs access to a process's corpus to get this process's details
        self.private_corpus.memberships.create(user=self.user, level=Role.Guest.value)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.user_img_process.id))

    def test_backwards_compatible_details(self):
        self.client.force_login(self.user)
        self.private_corpus.memberships.create(user=self.user, level=Role.Guest.value)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:import-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.user_img_process.id))

    def test_details_training_process(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-details', kwargs={'pk': self.training_process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            'activity_state': ActivityState.Disabled.value,
            'corpus': str(self.corpus.id),
            'element': None,
            'element_name_contains': None,
            'element_type': None,
            'files': [],
            'folder_type': None,
            'id': str(self.training_process.id),
            'load_children': False,
            'mode': ProcessMode.Training.value,
            'model_id': str(self.training_model.id),
            'name': None,
            'revision': None,
            'state': 'unscheduled',
            'template_id': None,
            'test_folder_id': str(self.test.id),
            'train_folder_id': str(self.train.id),
            'use_cache': False,
            'use_gpu': False,
            'validation_folder_id': str(self.val.id),
            'workflow': None,
        })

    def test_delete_requires_verified(self):
        user = User.objects.create(email='AAAAAAA')
        self.client.force_login(user)
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_delete_running(self):
        """
        It is not possible to delete a running import
        """
        self.client.force_login(self.user)
        self.elts_process.start()
        self.elts_process.workflow.tasks.update(state=State.Running)
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'__all__': ['Cannot delete a workflow while it is running']})

    def test_delete_activities_pending(self):
        """
        It is not possible to delete an import with non initialized activities
        """
        self.client.force_login(self.user)
        self.elts_process.start()
        self.elts_process.activity_state = ActivityState.Pending
        self.elts_process.save()
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Please wait activities to be initialized before deleting this process']}
        )

    def test_delete_repository_import_no_permission(self):
        """
        Deletion of a repository import requires to be admin on the repository
        """
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.repository_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_delete_corpus_import_no_permission(self):
        """
        A user cannot delete a process linked to a corpus he has no admin access to
        """
        self.client.force_login(self.user)
        self.assertFalse(self.user_img_process.corpus.memberships.filter(user=self.user).exists())
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {'detail': 'Not found.'})

    def test_delete_corpus_import_admin(self):
        """
        A user is allowed to delete a process if he has an admin right to its corpus.
        If no activities exists for this process, it is deleted directly.
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Process.DoesNotExist):
            self.elts_process.refresh_from_db()

    def test_delete_superuser(self):
        """
        A superuser is allowed to delete any process
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(10):
            response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Process.DoesNotExist):
            self.user_img_process.refresh_from_db()

    @patch('arkindex.project.triggers.process_tasks.process_delete.delay')
    def test_delete_corpus_import_async(self, delay_mock):
        """
        A process with activities is deleted in an async task and a HTTP_202_ACCEPTED is returned
        """
        self.client.force_login(self.user)
        self.elts_process.activities.create(
            element=self.corpus.elements.get(name='Volume 1'),
            worker_version=WorkerVersion.objects.get(worker__slug='reco'),
        )
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        self.assertEqual(delay_mock.call_count, 1)
        (process_arg,), kwargs = delay_mock.call_args
        self.assertEqual(process_arg.id, self.elts_process.id)
        self.assertEqual(kwargs, {'delete_activity': WorkerActivityState.Queued})

    @patch('arkindex.project.triggers.process_tasks.process_delete.delay')
    def test_delete_process_activities(self, delay_mock):
        """
        Deleting a process deletes the related activities with a queued state and
        sets the the FK on remaining activities to null
        Mock the async deletion task to run it directly in this test
        """
        from arkindex.process.tasks import process_delete
        delay_mock.side_effect = process_delete
        self.client.force_login(self.superuser)
        queued_activity = WorkerActivity.objects.create(
            element=self.corpus.elements.get(name='Volume 1'),
            process=self.elts_process,
            worker_version=WorkerVersion.objects.get(worker__slug='reco'),
            state=WorkerActivityState.Queued,
        )
        completed_activity = WorkerActivity.objects.create(
            element=self.corpus.elements.get(name='Volume 1, page 2r'),
            process=self.elts_process,
            worker_version=WorkerVersion.objects.get(worker__slug='reco'),
            state=WorkerActivityState.Processed,
        )
        with self.assertNumQueries(11):
            response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
            self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        with self.assertRaises(Process.DoesNotExist):
            self.elts_process.refresh_from_db()
        with self.assertRaises(WorkerActivity.DoesNotExist):
            queued_activity.refresh_from_db()
        with self.assertRaises(WorkerActivity.DoesNotExist):
            completed_activity.refresh_from_db()

    def test_update_process_requires_login(self):
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_update_process_requires_verified(self):
        user = User.objects.create(email="unchecked")
        self.client.force_login(user)
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_update_process_running(self):
        """
        A running process cannot be updated... (see following test)
        """
        self.client.force_login(self.user)
        self.elts_process.start()
        self.elts_process.workflow.tasks.update(state=State.Running)
        with self.assertNumQueries(7):
            response = self.client.patch(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'__all__': ['Cannot edit a workflow while it is running']})

    def test_update_process_running_only_name(self):
        """
        Only name attribute of a running process can be patched
        """
        self.client.force_login(self.user)
        self.elts_process.start()
        self.elts_process.workflow.tasks.update(state=State.Running)
        self.assertEqual(self.elts_process.name, None)
        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse('api:process-details', kwargs={'pk': self.elts_process.id}),
                {'name': 'newName'},
                format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.elts_process.refresh_from_db()
        self.assertEqual(self.elts_process.name, 'newName')

    def test_update_file_import_element(self):
        """
        A file import's element can be updated even while it is running
        """
        self.client.force_login(self.user)
        process = self.corpus.processes.create(mode=ProcessMode.Files, creator=self.user)
        process.start()
        process.workflow.tasks.update(state=State.Running)
        self.assertIsNone(self.elts_process.element)

        # Take any element with an image: this will cause extra queries to retrieve the Image and ImageServer
        element = self.corpus.elements.exclude(image_id=None).first()
        with self.assertNumQueries(13):
            response = self.client.patch(
                reverse('api:process-details', kwargs={'pk': process.id}),
                {'element_id': str(element.id)},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        self.assertEqual(process.element, element)

    def test_update_process_no_permission(self):
        """
        A user cannot update a process linked to a corpus he has no admin access to
        """
        self.client.force_login(self.user)
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        response = self.client.patch(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_update_process_repository_requires_admin(self):
        """
        Edition of a repository import requires to be admin on the repository
        """
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:process-details', kwargs={'pk': self.repository_process.id}),
            {'name': 'newName'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_update_process_wrong_mode(self):
        """
        Processes that are not of Workers mode cannot be updated, except the 'name' field
        """
        self.client.force_login(self.superuser)
        forbidden_modes = list(ProcessMode)
        forbidden_modes.remove(ProcessMode.Workers)
        # Local processes cannot be accessed by anyone because they have no corpus and no revision,
        # not even instance admins, so they cause a 404 instead of 400
        forbidden_modes.remove(ProcessMode.Local)
        # Tested separately; this process mode does not have a corpus, which causes other errors.
        forbidden_modes.remove(ProcessMode.Repository)

        for mode in forbidden_modes:
            with self.subTest(mode=mode):
                process = Process.objects.create(
                    mode=mode,
                    creator=self.user,
                    corpus=self.corpus,
                )

                response = self.client.put(
                    reverse('api:process-details', kwargs={'pk': process.id}),
                    {
                        'name': 'newName',
                        'template_id': None,
                        'element_name_contains': 'AAA',
                        'element_type': 'page',
                        'load_children': True
                    },
                    format='json'
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    response.json(),
                    {'__all__': ['Only processes of mode Workers can be updated']}
                )

    def test_update_process_repository(self):
        """
        Trying to update any field other than the `name` field on a Repository process is not allowed
        """
        self.client.force_login(self.superuser)
        process = Process.objects.create(
            creator=self.user,
            mode=ProcessMode.Repository,
            revision=self.rev,
        )
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:process-details', kwargs={'pk': process.id}),
                {
                    'name': 'newName',
                    'template_id': None,
                    'element_type': 'page',
                    'element_name_contains': 'AAA',
                    'load_children': True,
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'element_type': ['Object with slug=page does not exist.']},
        )

    def test_update_process_wrong_mode_patch(self):
        """
        Fields other than 'name' cannot be updated in processes that are not of Workers mode
        """
        self.client.force_login(self.superuser)
        forbidden_modes = list(ProcessMode)
        forbidden_modes.remove(ProcessMode.Workers)
        # Local processes cannot be accessed by anyone because they have no corpus and no revision,
        # not even instance admins, so they cause a 404 instead of 400
        forbidden_modes.remove(ProcessMode.Local)

        for mode in forbidden_modes:
            with self.subTest(mode=mode):
                process = Process(
                    mode=mode,
                    creator=self.user,
                    corpus=self.corpus,
                )
                if mode == ProcessMode.Repository:
                    process.corpus = None
                    process.revision = self.rev
                process.save()

                response = self.client.patch(
                    reverse('api:process-details', kwargs={'pk': process.id}),
                    {},
                    format='json',
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    response.json(),
                    {'__all__': ['Only processes of mode Workers can be updated']}
                )

    def test_update_process_only_name(self):
        """
        Whatever the process mode, except Local, name attribute (alone) can be updated
        """
        self.client.force_login(self.superuser)
        modes = list(ProcessMode)
        # Local processes cannot be accessed by anyone because they have no corpus and no revision,
        # not even instance admins, so they cause a 404
        modes.remove(ProcessMode.Local)

        for mode in modes:
            with self.subTest(mode=mode):
                process = Process(
                    mode=mode,
                    creator=self.user,
                    corpus=self.corpus,
                )
                if mode == ProcessMode.Repository:
                    process.corpus = None
                    process.revision = self.rev
                process.save()

                self.assertEqual(process.name, None)
                response = self.client.patch(
                    reverse('api:process-details', kwargs={'pk': process.id}),
                    {
                        'name': 'newName'
                    },
                    format='json'
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                process.refresh_from_db()
                self.assertEqual(process.name, 'newName')

    def test_update_process_corpus_no_write_right(self):
        self.client.force_login(self.user)
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        response = self.client.patch(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_update_process(self):
        """
        Process fields related to elements filtering can be updated
        """
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)
        self.assertTupleEqual(
            (process.name, process.name_contains, process.element_type, process.load_children),
            (None, None, None, False)
        )
        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse('api:process-details', kwargs={'pk': process.id}),
                {
                    'name': 'newName',
                    'element_name_contains': 'AAA',
                    'element_type': 'page',
                    'load_children': True,
                },
                format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        self.assertTupleEqual(
            (process.name, process.name_contains, process.element_type, process.load_children),
            ('newName', 'AAA', self.page_type, True)
        )

    def test_update_process_reset_fields(self):
        self.client.force_login(self.user)
        process = Process.objects.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
            creator=self.user,
            name='newName',
            name_contains='AAA',
            element_type=self.page_type,
            load_children=True
        )
        response = self.client.patch(
            reverse('api:process-details', kwargs={'pk': process.id}),
            {
                'name': None,
                'element_name_contains': None,
                'element_type': None,
                'load_children': False
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        self.assertTupleEqual(
            (process.name, process.name_contains, process.element_type, process.load_children),
            (None, None, None, False)
        )

    def test_update_process_wrong_type(self):
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)
        response = self.client.patch(
            reverse('api:process-details', kwargs={'pk': process.id}),
            {'element_type': 'unexisting_type'},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'element_type': ['Object with slug=unexisting_type does not exist.']}
        )

    def test_update_process_fields(self):
        """
        Ensure determined fields can be updated on a non started workers process.
        Editable fields are `name`, `name_contains`, `element_type`, and `load_children`.
        """
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)
        response = self.client.patch(
            reverse('api:process-details', kwargs={'pk': process.id}),
            {
                'id': str(uuid.uuid4()),
                'revision': str(uuid.uuid4()),
                'element': str(uuid.uuid4()),
                'workflow': str(uuid.uuid4()),
                'creator': str(uuid.uuid4()),
                'files': [1, 2, 3],
                'mode': 'iiif',
                'folder_type': 'register',
                'state': 'success',
                'name': 'newName',
                'element_name_contains': 'AAA',
                'element_type': 'page',
                'load_children': True,
                'use_cache': True,
                'use_gpu': False,
                'activity_state': ActivityState.Pending.value,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.json())
        self.assertEqual(response.json(), {
            'name': 'newName',
            'element_name_contains': 'AAA',
            'element_type': 'page',
            'load_children': True,
            'use_cache': False,
            'use_gpu': False,
            'activity_state': ActivityState.Disabled.value,
            'id': str(process.id),
            'corpus': str(self.corpus.id),
            'element': None,
            'files': [],
            'mode': 'workers',
            'revision': None,
            'state': 'unscheduled',
            'workflow': None,
            'template_id': None,
            'model_id': None,
            'train_folder_id': None,
            'validation_folder_id': None,
            'test_folder_id': None,
        })

    def test_retry_requires_login(self):
        response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retry_repository_process_no_right(self):
        """
        A user that is not the creator nor admin cannot restart a process that is not linked to any corpus
        """
        self.client.force_login(self.user)
        self.repository_process.start()
        self.repository_process.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.repository_process.state, State.Error)
        response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.repository_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have an admin access to this process.'})

    def test_retry_only_final(self):
        self.client.force_login(self.user)
        self.elts_process.start()
        parameters = [
            (State.Unscheduled, 'This workflow is already pending'),
            (State.Pending, 'This workflow is already pending'),
            (State.Running, 'This workflow is already running'),
            (State.Stopping, 'This workflow is stopping'),
        ]
        for state, message in parameters:
            with self.subTest(state=state):
                self.elts_process.workflow.tasks.all().update(state=state)
                self.assertEqual(self.elts_process.state, state)
                with self.assertNumQueries(7):
                    response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.elts_process.id}))
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), {'__all__': [message]})

    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    def test_retry(self, delay_mock):
        self.client.force_login(self.user)
        self.elts_process.start()
        self.elts_process.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.elts_process.state, State.Error)
        with self.assertNumQueries(16):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.elts_process.refresh_from_db()
        self.assertEqual(self.elts_process.state, State.Unscheduled)
        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    def test_retry_no_workflow(self, delay_mock):
        self.client.force_login(self.user)
        self.assertIsNone(self.elts_process.workflow)
        with self.assertNumQueries(18):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.elts_process.refresh_from_db()
        self.assertEqual(self.elts_process.state, State.Unscheduled)
        self.assertIsNotNone(self.elts_process.workflow)
        self.assertFalse(delay_mock.called)

    def test_retry_repo_disabled(self):
        self.repository_process.revision = self.rev
        self.repository_process.save()
        self.repository_process.start()
        self.repository_process.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.repository_process.state, State.Error)
        self.creds.delete()
        # Allow the user to do the retry
        self.repo.memberships.filter(user=self.user).update(level=Role.Admin.value)
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.repository_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ['Git repository does not have any valid credentials'])

    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    def test_retry_initialize_activity(self, delay_mock):
        """
        Retrying a process should re-initialize the WorkerActivities if they are enabled
        """
        self.client.force_login(self.user)
        self.workers_process.start()
        self.workers_process.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.workers_process.state, State.Error)

        self.workers_process.activity_state = ActivityState.Error
        self.workers_process.save()

        with self.assertNumQueries(25):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.workers_process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.state, State.Unscheduled)
        self.assertEqual(self.workers_process.activity_state, ActivityState.Pending)
        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(self.workers_process))

    def test_from_files_requires_login(self):
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_from_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.mode, ProcessMode.Files)
        self.assertListEqual(list(process.files.all()), [self.img_df])
        self.assertEqual(process.folder_type.slug, 'volume')
        self.assertEqual(process.element_type.slug, 'page')
        self.assertIsNone(process.element)

    def test_from_files_pdf(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.pdf_df.id)],
            'mode': 'files',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.mode, ProcessMode.Files)
        self.assertListEqual(list(process.files.all()), [self.pdf_df])
        self.assertIsNone(process.element)

    def test_from_files_image_and_pdf(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.pdf_df.id), str(self.img_df.id)],
            'mode': 'files',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.mode, ProcessMode.Files)
        self.assertListEqual(list(process.files.all()), [self.img_df, self.pdf_df])
        self.assertIsNone(process.element)

    @override_settings(
        PONOS_RECIPE={'env': {'ARKINDEX_API_TOKEN': 'testToken'}},
        ARKINDEX_TASKS_IMAGE='registry.teklia.com/tasks',
    )
    def test_from_files_iiif(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(34), self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            response = self.client.post(reverse('api:files-process'), {
                'files': [str(self.iiif_df.id)],
                'mode': 'iiif',
                'folder_type': 'volume',
                'element_type': 'page',
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        process = Process.objects.get(id=data['id'])

        self.assertEqual(process.mode, ProcessMode.IIIF)
        self.assertListEqual(list(process.files.all()), [self.iiif_df])
        self.assertEqual(process.folder_type.slug, 'volume')
        self.assertEqual(process.element_type.slug, 'page')
        self.assertIsNone(process.element)
        self.assertEqual(process.creator, self.user)
        self.assertEqual(process.corpus, self.corpus)
        self.assertIsNone(process.revision)

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.import_worker_version)
        self.assertListEqual(worker_run.parents, [])
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)

        self.assertIsNotNone(process.workflow)
        recipe = yaml.safe_load(process.workflow.recipe)
        self.maxDiff = None
        self.assertDictEqual(recipe, {
            'env': {
                'ARKINDEX_API_TOKEN': 'testToken',
                'ARKINDEX_PROCESS_ID': str(process.id),
                'ARKINDEX_CORPUS_ID': str(self.corpus.id),
                'ARKINDEX_WORKER_RUN_ID': str(worker_run.id),
            },
            'tasks': {
                'import': {
                    'image': 'registry.teklia.com/tasks',
                    'command': f'python -m arkindex_tasks.import_iiif.process {process.id}',
                },
                'thumbnails': {
                    'image': 'registry.teklia.com/tasks',
                    'command': 'python3 -m arkindex_tasks.generate_thumbnails /data/import/elements.json',
                    'parents': ['import'],
                },
            },
        })

    def test_from_files_iiif_with_json_charset(self):
        self.iiif_df.content_type = 'application/json;charset=utf-8'
        self.assertEqual(self.iiif_df.content_type, 'application/json;charset=utf-8')

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.mode, ProcessMode.IIIF)
        self.assertListEqual(list(process.files.all()), [self.iiif_df])
        self.assertEqual(process.folder_type.slug, 'volume')
        self.assertEqual(process.element_type.slug, 'page')
        self.assertIsNone(process.element)

    def test_from_files_iiif_with_json_profile_and_charset(self):
        self.iiif_df.content_type = 'application/ld+json;profile="http://iiif.io/api/presentation/2/context.json";charset=utf-8'
        self.assertEqual(self.iiif_df.content_type, 'application/ld+json;profile="http://iiif.io/api/presentation/2/context.json";charset=utf-8')

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.mode, ProcessMode.IIIF)
        self.assertListEqual(list(process.files.all()), [self.iiif_df])
        self.assertEqual(process.folder_type.slug, 'volume')
        self.assertEqual(process.element_type.slug, 'page')
        self.assertIsNone(process.element)

    def test_from_files_iiif_requires_folder_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['IIIF imports require both folder_type and element_type']
        })

    def test_from_files_invalid_mode(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'repository',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_iiif_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.pdf_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'iiif',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'non_field_errors': ['IIIF imports can only import IIIF documents']})

    def test_from_files_files_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.iiif_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'files',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'non_field_errors': ['File imports can only import images or PDF documents']})

    def test_from_files_folder_id(self):
        self.client.force_login(self.user)
        volume = self.corpus.elements.get(name='Volume 1')
        response = self.client.post(
            reverse('api:files-process'),
            {
                'files': [str(self.pdf_df.id)],
                'mode': 'files',
                'folder_id': str(volume.id),
                'element_type': 'page',
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.element, volume)

    def test_from_files_no_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_start_process_requires_login(self):
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(self.user_img_process.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_start_process_invalid_process_id(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_start_process_wrong_process_mode(self):
        self.assertNotEqual(self.user_img_process.mode, ProcessMode.Workers)
        # grant an admin access to this process
        self.user_img_process.corpus.memberships.create(user=self.user, level=Role.Admin.value)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(self.user_img_process.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'__all__': ['Only a Process with Workers mode and not already launched can be started later on']}
        )

    def test_start_process_process_already_started(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process2.workflow = Workflow.objects.create(farm_id=get_default_farm_id(), recipe=RECIPE)
        process2.save()
        self.assertIsNotNone(process2.workflow)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(process2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'__all__': ['Only a Process with Workers mode and not already launched can be started later on']}
        )

    def test_start_process_without_required_model(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process2.worker_runs.create(version=self.version_with_model, parents=[], configuration=None)
        process2.save()

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(process2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {"model_version": [f"The following workers require a model version and none was set: {[self.version_with_model.worker.name]}"]}
        )

    def test_start_process_with_required_model(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        run = process2.worker_runs.create(version=self.version_with_model, parents=[], configuration=None)
        run.model_version = self.model_version_1
        run.save()
        self.assertIsNone(process2.workflow)

        self.client.force_login(self.user)
        with self.assertNumQueries(25):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], str(process2.id))

    def test_start_process_empty(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        self.assertIsNone(process2.workflow)

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'__all__': ['The process must either use thumbnail generation or have worker runs.']},
        )

    def test_start_process_unavailable_worker_version(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process2.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.recognizer.state = WorkerVersionState.Error
        self.recognizer.save()
        self.assertIsNone(process2.workflow)

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'version': 'This process contains one or more unavailable worker versions and cannot be started.'},
        )

    def test_start_process_unavailable_model_version(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process2.worker_runs.create(version=self.recognizer, parents=[], configuration=None, model_version=self.model_version_1)
        self.model_version_1.state = ModelVersionState.Error
        self.model_version_1.save()
        self.assertIsNone(process2.workflow)

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'model_version': ['This process contains one or more unavailable model versions and cannot be started.']},
        )

    def test_start_process(self):
        """
        A user can start a process with no parameters.
        Default chunks, thumbnails and farm are used. Nor cache or workers activity is set.
        """
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process2.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertIsNone(process2.workflow)

        self.client.force_login(self.user)
        with self.assertNumQueries(25):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], str(process2.id))
        process2.refresh_from_db()
        self.assertEqual(process2.state, State.Unscheduled)
        self.assertIsNotNone(process2.workflow)
        # Ensure default parameters are used
        self.assertEqual(process2.workflow.farm_id, get_default_farm_id())
        self.assertEqual(process2.workflow.tasks.count(), 2)
        task1, task2 = process2.workflow.tasks.order_by('slug')
        self.assertEqual(task1.slug, 'initialisation')
        self.assertEqual(task2.slug, f'reco_{str(self.recognizer.id)[:6]}')
        self.assertIn('--chunks-number 1', task1.command)

    def test_start_process_select_farm_id(self):
        """
        A user can specify a ponos farm to use for a workflow
        """
        get_default_farm_id()
        barley_farm = Farm.objects.create(name='Barley farm')
        self.assertNotEqual(get_default_farm_id(), barley_farm.id)
        workers_process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        workers_process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.client.force_login(self.user)
        with self.assertNumQueries(25):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(workers_process.id)}),
                {'farm': str(barley_farm.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        workers_process.refresh_from_db()
        self.assertEqual(workers_process.state, State.Unscheduled)
        self.assertEqual(workers_process.workflow.farm_id, barley_farm.id)

    def test_start_process_wrong_farm_id(self):
        """
        Handle a payload with a wrong farm ID starting a process
        """
        workers_process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        self.client.force_login(self.user)
        wrong_farm_id = uuid.uuid4()
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(workers_process.id)}),
                {'farm': str(wrong_farm_id)}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'farm': [f'Invalid pk "{wrong_farm_id}" - object does not exist.']})

    def test_start_process_params_validation(self):
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        self.client.force_login(self.user)
        wrong_params_checks = [
            ({'chunks': 0}, {'chunks': ['Ensure this value is greater than or equal to 1.']}),
            ({'chunks': 20}, {'chunks': ['Ensure this value is less than or equal to 10.']}),
            ({'chunks': 'max'}, {'chunks': ['A valid integer is required.']}),
            ({'thumbnails': 'gloubiboulga'}, {'thumbnails': ['Must be a valid boolean.']})
        ]
        for (params, check) in wrong_params_checks:
            response = self.client.post(reverse('api:process-start', kwargs={'pk': str(process.id)}), params)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(response.json(), check)

    @override_settings(MAX_CHUNKS=42)
    def test_start_process_serializer_chunks_config(self):
        """
        Assert the start process serializer use configuration max chunks
        """
        # Reloading serializer classes is required to update fields attributes from settings
        from arkindex.process.serializers import imports
        imports = importlib.reload(imports)

        params_checks = [
            ({'chunks': 42}, True, None),
            ({'chunks': 43}, False, {'chunks': ['Ensure this value is less than or equal to 42.']}),
        ]
        for (params, is_valid, message) in params_checks:
            serializer = imports.StartProcessSerializer(data=params)
            if not is_valid:
                with self.assertRaisesRegex(ValidationError, 'Ensure this value is less than or equal to 42.'):
                    self.assertEqual(serializer.is_valid(raise_exception=True))
            else:
                self.assertEqual(serializer.is_valid(), True)

    @patch('arkindex.process.models.Process.worker_runs')
    def test_start_process_workflow_parameters(self, worker_runs_mock):
        """
        It should be possible to pass chunks and thumbnails parameters when starting a workflow
        """
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        # Add a worker run to this process
        run_mock = MagicMock()
        run_mock.version.slug = 'my_worker'
        run_mock.build_task_recipe.return_value = {'image': ''}
        # Cheat to mock a query on process worker runs
        worker_runs_mock.select_related.return_value = worker_runs_mock
        worker_runs_mock.using.return_value = [run_mock]

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(process.id)}),
            {'chunks': 3, 'thumbnails': 'true'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        recipe_dump = yaml.safe_load(process.workflow.recipe)

        self.assertEqual(
            list(recipe_dump['tasks'].keys()),
            ['initialisation', 'my_worker_1', 'my_worker_2', 'my_worker_3', 'thumbnails_1', 'thumbnails_2', 'thumbnails_3']
        )

    @patch('arkindex.process.models.Process.worker_runs')
    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    def test_start_process_options_requires_workers(self, activities_delay_mock, worker_runs_mock):
        """
        Cache and worker activity options can be triggered when starting a process
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process.id)}),
                {'use_cache': 'true', 'worker_activity': 'true', 'use_gpu': 'true'}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            '__all__': ['The process must either use thumbnail generation or have worker runs.'],
            'use_cache': 'The process must have workers attached to use cached results.',
            'worker_activity': 'The process must have workers attached to handle their activity.',
            'use_gpu': 'The process must have workers attached to use GPUs.'
        })
        process.refresh_from_db()
        self.assertFalse(process.use_cache)
        self.assertEqual(process.activity_state, ActivityState.Disabled)

    @patch('arkindex.process.models.Process.worker_runs')
    @patch('arkindex.process.models.Process.versions')
    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    def test_start_process_options(self, activities_delay_mock, versions_mock, worker_runs_mock):
        """
        Cache and worker activity options requires the process to have workers attached
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        self.assertFalse(process.use_cache)
        self.assertEqual(process.activity_state, ActivityState.Disabled)

        # Add a worker run to this process
        run_mock = MagicMock()
        run_mock.version = self.version_gpu
        run_mock.model_version_id = None
        run_mock.build_task_recipe.return_value = {'image': ''}
        # Cheat to mock a query on process worker runs
        worker_runs_mock.all.return_value = [run_mock]

        self.client.force_login(self.user)
        with self.assertNumQueries(18):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process.id)}),
                {'use_cache': 'true', 'worker_activity': 'true', 'use_gpu': 'true'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        self.assertTrue(process.use_cache)
        self.assertTrue(process.use_gpu)
        # The process activity state is set to Pending while activities are created in an async job
        self.assertEqual(process.activity_state, ActivityState.Pending)
        self.assertEqual(activities_delay_mock.call_count, 1)

    @override_settings(
        PONOS_RECIPE={'env': {'ARKINDEX_API_TOKEN': 'testToken'}},
        ARKINDEX_TASKS_IMAGE='registry.teklia.com/tasks',
    )
    def test_pdf_import_build_workflow_recipe(self):
        """
        Build a workflow for a PDF import with a worker version defined in settings
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Files,
        )
        process.files.create(name='A PDF file', size=10e6, content_type='pdf', corpus=self.corpus)

        with self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            workflow = process.build_workflow()

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.import_worker_version)
        self.assertListEqual(worker_run.parents, [])
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)

        recipe = yaml.safe_load(workflow.recipe)
        self.assertDictEqual(recipe, {
            'env': {
                'ARKINDEX_API_TOKEN': 'testToken',
                'ARKINDEX_PROCESS_ID': str(process.id),
                'ARKINDEX_CORPUS_ID': str(self.corpus.id),
                'ARKINDEX_WORKER_RUN_ID': str(worker_run.id),
            },
            'tasks': {
                'import': {
                    'command': f'python -m arkindex_tasks.import_files {process.id}',
                    'image': 'registry.teklia.com/tasks'
                }
            }
        })

    @override_settings(PONOS_RECIPE={'env': {'ARKINDEX_API_TOKEN': 'testToken'}})
    @override_settings(ARKINDEX_TASKS_IMAGE='registry.teklia.com/tasks')
    @override_settings(PUBLIC_HOSTNAME='https://arkindex.localhost')
    def test_worker_run_model_version_build_workflow_recipe(self):
        """
        Build a workflow for a PDF import with a worker version defined in settings
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
        )
        run = process.worker_runs.create(version=self.version_with_model, parents=[], configuration=None)
        run.model_version = self.model_version_1
        run.save()
        workflow = process.build_workflow()
        recipe = yaml.safe_load(workflow.recipe)
        self.assertCountEqual(recipe.keys(), ['env', 'tasks'])
        self.assertDictEqual(recipe['env'], {
            'ARKINDEX_API_TOKEN': 'testToken',
            'ARKINDEX_PROCESS_ID': str(process.id),
            'ARKINDEX_CORPUS_ID': str(self.corpus.id),
        })

        self.assertCountEqual(
            recipe['tasks'].values(),
            [
                {
                    'command': f'python -m arkindex_tasks.init_elements {process.id} --chunks-number 1',
                    'image': 'registry.teklia.com/tasks'
                },
                {
                    'image': run.version.docker_image_iid or run.version.docker_image_name,
                    'command': run.version.docker_command,
                    'artifact': str(run.version.docker_image_id),
                    'parents': ['initialisation'],
                    'shm_size': None,
                    'env': {
                        "TASK_ELEMENTS": "/data/initialisation/elements.json",
                        "WORKER_VERSION_ID": str(run.version.id),
                        "ARKINDEX_WORKER_RUN_ID": str(run.id),
                    },
                    "extra_files": {
                        "model": "https://arkindex.localhost" + reverse('api:model-version-download', kwargs={'pk': run.model_version.id}) + f"?token={run.model_version.build_authentication_token_hash()}"
                    }
                }
            ]
        )

    def test_clear_process(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        process.refresh_from_db()
        self.assertEqual(process.worker_runs.count(), 0)

    def test_clear_process_requires_login(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)

        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_clear_process_requires_verified(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)

        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_clear_process_does_not_exist(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)

        fake = uuid.uuid4()

        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(fake)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_clear_process_not_unscheduled(self):
        """
        Cannot clear a process that has already started
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)
        process.start()
        process.workflow.tasks.update(state=State.Running)
        self.assertEqual(process.state, State.Running)

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'__all__': ['A process can only be cleared before getting started.']})

    def test_clear_process_unscheduled_workflow(self):
        """
        Cannot clear a process that has a workflow, even unscheduled
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)
        process.start()
        self.assertEqual(process.state, State.Unscheduled)

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'__all__': ['A process can only be cleared before getting started.']})

    def test_clear_process_requires_permissions(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)

        user2 = User.objects.create_user('email@mail.com', 'bob')
        user2.verified_email = True
        user2.save()

        self.corpus.memberships.create(user=user2, level=Role.Contributor.value)
        self.client.force_login(user2)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})
