from unittest.mock import patch

import yaml
from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import State
from arkindex.process.models import Process, ProcessMode, Repository, Revision, Worker, WorkerType, WorkerVersion
from arkindex.project.default_corpus import DEFAULT_TRANSKRIBUS_TYPES
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


@override_settings(TRANSKRIBUS_EMAIL="arkindex@teklia.com", TRANSKRIBUS_PASSWORD=None)
class TestTranskribusImport(FixtureAPITestCase):
    """
    Test transkribus import
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        User.objects.update(transkribus_email="nope@nope.fr")
        cls.repo = Repository.objects.create(url='http://fakery')
        cls.tr_type = WorkerType.objects.get(slug='import')
        cls.transkribus_worker = Worker.objects.create(
            repository=cls.repo,
            name='Transkribus Import',
            slug='transkribus_import',
            type=cls.tr_type
        )
        cls.transkribus_rev = Revision.objects.create(
            hash='1234',
            message='commit commit',
            author='Wolpertinger',
            repo=cls.repo,
        )
        cls.transkribus_worker_version = WorkerVersion.objects.create(
            worker=cls.transkribus_worker,
            revision=cls.transkribus_rev,
            configuration={}
        )

    @override_settings(ARKINDEX_FEATURES={"transkribus": True})
    def test_requires_login(self):
        response = self.client.post(reverse("api:import-transkribus"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"transkribus": False})
    def test_requires_flag(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:import-transkribus"), {
            "collection_id": "12345",
        }, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ["Transkribus import is unavailable due to the transkribus feature being disabled."])

    @override_settings(ARKINDEX_FEATURES={"transkribus": True})
    def test_requires_transkribus_email(self):
        User.objects.update(transkribus_email=None)
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:import-transkribus"), {
            "collection_id": "12345",
        }, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"__all__": ["You have not registered your transkribus email"]})

    @override_settings(ARKINDEX_FEATURES={"transkribus": True})
    @patch("transkribus.TranskribusAPI.list_user_collection")
    def test_arkindex_has_not_access(self, mock_transkribus):
        # Not a mistake: Transkribus client raises `Exception` directly when it runs out of retries
        mock_transkribus.side_effect = Exception("401 Unauthorized")
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:import-transkribus"), {
            "collection_id": "12345",
        }, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"collection_id" : ["User arkindex@teklia.com is not a member of the collection 12345"]})

    @override_settings(ARKINDEX_FEATURES={"transkribus": True})
    @patch("transkribus.TranskribusAPI.list_user_collection")
    def test_user_has_not_access(self, mock_transkribus):
        mock_transkribus.return_value = [{"email": "arkindex@teklia.com"}]

        self.client.force_login(self.user)
        response = self.client.post(reverse("api:import-transkribus"), {
            "collection_id": "12345",
        }, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"collection_id": ["User nope@nope.fr is not a member of the collection 12345"]})

    @override_settings(
        PONOS_RECIPE={},
        ARKINDEX_FEATURES={"transkribus": True},
        TRANSKRIBUS_EMAIL="arkindex@teklia.com",
        TRANSKRIBUS_PASSWORD="averysecretpassword",
        ARKINDEX_TASKS_IMAGE='registry.teklia.com/tasks',
    )
    @patch("arkindex.process.serializers.imports.TranskribusAPI")
    def test_create_import(self, mock_transkribus):
        mock_transkribus.return_value.list_user_collection.return_value = [{"email": "nope@nope.fr"}, {"email": "arkindex@teklia.com"}]
        self.client.force_login(self.user)

        with self.settings(TRANSKRIBUS_WORKER_VERSION=self.transkribus_worker_version.id):
            response = self.client.post(reverse("api:import-transkribus"), {
                "collection_id": "12345",
            }, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        process = Process.objects.get(id=data["id"])

        self.assertEqual(process.mode, ProcessMode.Transkribus)
        self.assertEqual(process.collection_id, 12345)
        self.assertEqual(process.state, State.Unscheduled)

        corpus = process.corpus
        self.assertEqual(corpus.name, "Transkribus collection n°12345")
        self.assertEqual(corpus.description, "")
        self.assertEqual(corpus.public, False)
        # Assert defaults types are set on the new corpus
        self.assertCountEqual(
            list(corpus.types.values(
                "slug",
                "display_name",
                "folder",
            )),
            [{
                "folder": False,
                **values
            } for values in DEFAULT_TRANSKRIBUS_TYPES]
        )

        right = corpus.memberships.get(user=self.user)
        self.assertTrue(right.level >= Role.Admin.value)

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.transkribus_worker_version)
        self.assertListEqual(worker_run.parents, [])
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)

        self.assertIsNotNone(process.workflow)
        recipe = yaml.safe_load(process.workflow.recipe)
        self.assertDictEqual(recipe, {
            'env': {
                'ARKINDEX_PROCESS_ID': str(process.id),
                'ARKINDEX_CORPUS_ID': str(corpus.id),
                'ARKINDEX_WORKER_RUN_ID': str(worker_run.id),
                'TRANSKRIBUS_EMAIL': "arkindex@teklia.com",
                'TRANSKRIBUS_PASSWORD': "averysecretpassword",
            },
            'tasks': {
                'export_transkribus': {
                    'command': 'python -m arkindex_tasks.export_transkribus 12345',
                    'image': 'registry.teklia.com/tasks',
                },
                'import_arkindex': {
                    'command': f'python -m arkindex_tasks.import_transkribus --job-path /data/export_transkribus/transkribus_export_job.json --corpus {corpus.id}',
                    'image': 'registry.teklia.com/tasks',
                    'parents': ['export_transkribus']
                },
                'thumbnails': {
                    'command': 'python3 -m arkindex_tasks.generate_thumbnails /data/import_arkindex/elements.json',
                    'image': 'registry.teklia.com/tasks',
                    'parents': ['import_arkindex']
                }
            }
        })

    @override_settings(
        PONOS_RECIPE={},
        ARKINDEX_FEATURES={'transkribus': True},
    )
    @patch("transkribus.TranskribusAPI.list_user_collection")
    def test_corpus_worker_version_list_transkribus(self, mock_transkribus):
        mock_transkribus.return_value = [{"email": "nope@nope.fr"}, {"email": "arkindex@teklia.com"}]
        self.client.force_login(self.user)
        with self.settings(TRANSKRIBUS_WORKER_VERSION=self.transkribus_worker_version.id):
            response = self.client.post(reverse("api:import-transkribus"), {
                "collection_id": "12345",
            }, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.mode, ProcessMode.Transkribus)
        corpus = process.corpus
        with self.assertNumQueries(10):
            response = self.client.get(reverse('api:corpus-versions', kwargs={'pk': corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(len(response.json()['results']), 1)
        self.assertEqual(response.json()['results'][0]['id'], str(self.transkribus_worker_version.id))
