import itertools
import uuid
from datetime import datetime, timedelta, timezone

from django.db.models.functions import Now
from django.urls import reverse
from rest_framework import status

import pgtrigger
from arkindex.documents.models import Corpus
from arkindex.process.models import (
    Process,
    ProcessMode,
    WorkerActivity,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerVersion,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestWorkersActivity(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.version_1 = WorkerVersion.objects.get(worker__slug='reco')
        cls.version_2 = WorkerVersion.objects.get(worker__slug='dla')
        cls.configuration_1 = WorkerConfiguration.objects.create(name='some_config', configuration={'aa': 'bb'}, worker=cls.version_1.worker)
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.process_1 = Process.objects.first()
        cls.process_2 = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            corpus=cls.corpus,
        )

        # Generate worker activities
        WorkerActivity.objects.bulk_create([
            *(
                WorkerActivity(
                    element_id=elt.id,
                    state=state,
                    worker_version_id=cls.version_1.id,
                    configuration_id=cls.configuration_1.id,
                    process_id=cls.process_1.id,
                    started=datetime.now(timezone.utc),
                ) for elt, state in zip(cls.corpus.elements.all(), itertools.cycle(WorkerActivityState))
            ), *(
                WorkerActivity(
                    element_id=elt.id,
                    state=WorkerActivityState.Processed.value,
                    worker_version_id=cls.version_2.id,
                    process_id=cls.process_2.id,
                ) for elt in cls.corpus.elements.all()
            )
        ])

        # Django's auto_now_add uses Python's datetime.now() and calls it once for each field.
        # This means there can be a random delay of a few microseconds between `created` and `updated` on each activity,
        # so we manually update all the activities to make them use the SQL NOW() function instead.
        # Setting `created=Now(), updated=Now()` on the WorkerActivity() above does not work because `auto_now_add`
        # is executed at the last moment by the ORM, right as it is about to insert, and ignores any previously defined values.
        # Since we have triggers to automatically update `updated` and prevent manual updates, we have to temporarily disable them.
        with pgtrigger.ignore('process.WorkerActivity:read_only_workeractivity_updated', 'process.WorkerActivity:update_workeractivity_updated'):
            WorkerActivity.objects.filter(element__corpus_id=cls.corpus.id).update(created=Now(), updated=Now(), started=Now())

        cls.error, cls.processed, cls.queued, cls.started = [
            WorkerActivity.objects.filter(
                element__corpus_id=cls.corpus.id,
                worker_version_id=cls.version_1.id,
                state=state
            ).count()
            for state in [
                WorkerActivityState.Error,
                WorkerActivityState.Processed,
                WorkerActivityState.Queued,
                WorkerActivityState.Started
            ]
        ]

    def test_corpus_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse('api:corpus-activity-stats', kwargs={'corpus': str(self.corpus.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_corpus_private(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-activity-stats', kwargs={'corpus': str(self.private_corpus.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have guest access to this corpus.'})

    def test_corpus_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse('api:corpus-activity-stats', kwargs={'corpus': str(uuid.uuid4())})
            )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_corpus_empty(self):
        """
        Handle a corpus that has absolutely no activity
        A user with a guest access can retrieve statistics
        """
        user = User.objects.create_user('user42@test.test', 'abcd')
        user.verified_email = True
        user.save()
        self.private_corpus.memberships.create(user=user, level=Role.Guest.value)
        self.client.force_login(user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-activity-stats', kwargs={'corpus': str(self.private_corpus.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [])

    def test_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-activity-stats', kwargs={'corpus': str(self.corpus.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(response.json(), [
            {
                'worker_version_id': str(self.version_1.id),
                'configuration_id': str(self.configuration_1.id),
                'queued': self.queued,
                'started': self.started,
                'processed': self.processed,
                'error': self.error,
                'average_processed_time': '00:00:00',
                'min_processed_time': '00:00:00',
                'max_processed_time': '00:00:00',
                'median_processed_time': '00:00:00',
            }, {
                'worker_version_id': str(self.version_2.id),
                'configuration_id': None,
                'queued': 0,
                'started': 0,
                'processed': self.corpus.elements.count(),
                'error': 0,
                'average_processed_time': '00:00:00',
                'min_processed_time': '00:00:00',
                'max_processed_time': '00:00:00',
                'median_processed_time': '00:00:00',
            }
        ])

    def test_corpus_timing(self):
        # Grab 4 processed WorkerActivities, and delete all the other processed activities.
        processed_activities = WorkerActivity.objects.filter(
            element__corpus_id=self.corpus.id,
            worker_version_id=self.version_1.id,
            state=WorkerActivityState.Processed,
        ).order_by('id')
        self.assertGreaterEqual(processed_activities.count(), 4)
        # id__in is necessary because Django does not support LIMIT/OFFSET on a DELETE
        WorkerActivity.objects.filter(id__in=processed_activities[4:].values('id')).delete()
        # Set a fixed start time so we can edit `updated` to change the processed time
        processed_activities.update(started=datetime(2012, 12, 12, 12, 12, 12, 12, timezone.utc))
        act1, act2, act3, act4 = processed_activities

        # With 4 activities, we can set some simple values to test the min, max, average and median.
        # With an even number, we can ensure that the median processed time can be computed between two values
        # and will not always be an existing value in the set.
        act1.updated = act1.started + timedelta(minutes=1)
        act2.updated = act2.started + timedelta(minutes=2)
        act3.updated = act3.started + timedelta(minutes=3)
        act4.updated = act4.started + timedelta(minutes=10)

        # Disable the Postgres triggers so we can force our updated times
        with pgtrigger.ignore('process.WorkerActivity:read_only_workeractivity_updated', 'process.WorkerActivity:update_workeractivity_updated'):
            WorkerActivity.objects.bulk_update([act1, act2, act3, act4], ['updated'])

        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-activity-stats', kwargs={'corpus': str(self.corpus.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertCountEqual(response.json(), [
            {
                'worker_version_id': str(self.version_1.id),
                'configuration_id': str(self.configuration_1.id),
                'queued': self.queued,
                'started': self.started,
                'processed': 4,
                'error': self.error,
                'average_processed_time': '00:04:00',
                'min_processed_time': '00:01:00',
                'max_processed_time': '00:10:00',
                'median_processed_time': '00:02:30',
            }, {
                'worker_version_id': str(self.version_2.id),
                'configuration_id': None,
                'queued': 0,
                'started': 0,
                'processed': self.corpus.elements.count(),
                'error': 0,
                'average_processed_time': '00:00:00',
                'min_processed_time': '00:00:00',
                'max_processed_time': '00:00:00',
                'median_processed_time': '00:00:00',
            }
        ])

    def test_corpus_timing_no_processed(self):
        """
        When there are no processed activities, all the aggregate functions will return `None`
        """
        WorkerActivity.objects.filter(state=WorkerActivityState.Processed).delete()
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-activity-stats', kwargs={'corpus': str(self.corpus.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(response.json(), [
            {
                'worker_version_id': str(self.version_1.id),
                'configuration_id': str(self.configuration_1.id),
                'queued': self.queued,
                'started': self.started,
                'processed': 0,
                'error': self.error,
                'average_processed_time': None,
                'min_processed_time': None,
                'max_processed_time': None,
                'median_processed_time': None,
            }
        ])

    def test_process_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse('api:process-activity-stats', kwargs={'pk': str(self.process_1.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_process_private(self):
        # An inaccessible project
        self.process_1.corpus = self.private_corpus
        self.process_1.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:process-activity-stats', kwargs={'pk': str(self.process_1.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_process_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse('api:process-activity-stats', kwargs={'pk': str(uuid.uuid4())})
            )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_process(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(
                reverse('api:process-activity-stats', kwargs={'pk': str(self.process_1.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [
            {
                'worker_version_id': str(self.version_1.id),
                'configuration_id': str(self.configuration_1.id),
                'queued': self.queued,
                'started': self.started,
                'processed': self.processed,
                'error': self.error,
                'average_processed_time': '00:00:00',
                'min_processed_time': '00:00:00',
                'max_processed_time': '00:00:00',
                'median_processed_time': '00:00:00',
            }
        ])
