from django.core.exceptions import ValidationError
from django.core.management.base import CommandError
from django.db.models import Model

from arkindex.documents.models import Corpus, Element
from arkindex.process.models import Process, Repository, WorkerVersion
from arkindex.users.models import User


class ModelArgument(object):
    model = None
    text_search_field = 'name'
    text_search_lookup = 'icontains'

    def __init__(self, many=False, **filters):
        assert issubclass(self.model, Model)
        self.many = many
        self.filters = filters

    def __call__(self, arg):
        qs = self.model.objects.filter(**self.filters)
        try:
            return qs.get(pk=arg)
        except (ValueError, ValidationError, self.model.DoesNotExist):
            pass

        try:
            return self.text_search(qs, arg)
        except self.model.DoesNotExist:
            raise CommandError('{} "{}" does not exist'.format(self.model.__name__, arg))
        except self.model.MultipleObjectsReturned:
            raise CommandError('"{}" matches multiple {} instances'.format(arg, self.model.__name__))

    def text_search(self, qs, arg):
        if not self.text_search_field:
            raise self.model.DoesNotExist
        text_filter = {f'{self.text_search_field}__{self.text_search_lookup}': arg}
        if self.many:
            return qs.filter(**text_filter)
        else:
            return qs.get(**text_filter)


class CorpusArgument(ModelArgument):
    model = Corpus


class ElementArgument(ModelArgument):
    model = Element


class ProcessArgument(ModelArgument):
    model = Process
    text_search_field = None


class RepositoryArgument(ModelArgument):
    model = Repository
    text_search_field = 'url'


class UserArgument(ModelArgument):
    model = User
    text_search_field = 'email'
    text_search_lookup = 'exact'


class WorkerVersionArgument(ModelArgument):
    model = WorkerVersion
    text_search_field = None
