import re

from drf_spectacular.drainage import warn
from drf_spectacular.settings import spectacular_settings

# Allow alphanumeric characters, except at the end of an operation ID:
# this is likely to be a suffix that Spectacular added by itself due to conflicting IDs.
OPERATION_ID_REGEX = re.compile(r'^[a-zA-Z0-9]*[a-zA-Z]$')


def operation_id_pascal_case(result, **kwargs):
    """
    Transform camelCased operation IDs into PascalCase
    """
    for methods in result['paths'].values():
        for method in methods.values():
            # Ensure there are no 'weird' operation IDs; for example, Spectacular can default to
            # snake_cased operation IDs, or cause duplicated operation IDs to have number suffixes
            if not OPERATION_ID_REGEX.search(method['operationId']):
                warn(f'Non-alphabetic characters found in operation ID {method["operationId"]}')

            method['operationId'] = method['operationId'][0].upper() + method['operationId'][1:]
    return result


def check_tags(result, **kwargs):
    """
    Check endpoint tags to require them to be defined in settings,
    and require all endpoints to have a tag
    """
    known_tags = set(tag['name'] for tag in spectacular_settings.TAGS)
    for methods in result['paths'].values():
        for method in methods.values():
            tags = method.get('tags')

            if not tags:
                warn(f'Endpoint {method["operationId"]} does not have any assigned tags')
                continue

            unknown_tags = set(tags) - known_tags
            if unknown_tags:
                warn(f'Endpoint {method["operationId"]} uses tags not defined in the project settings: {unknown_tags}')

    return result


def request_body_x_name(result, **kwargs):
    """
    Adds a `x-name` on request bodies to allow them to be used in APIStar
    with the `body={}` keyword argument
    """
    for methods in result['paths'].values():
        for method in methods.values():
            if method.get('requestBody'):
                method['requestBody']['x-name'] = 'body'

    return result
