from typing import Optional

from django_rq import get_connection
from django_rq.queues import DjangoRQ
from rq.compat import as_text, decode_redis_hash
from rq.exceptions import NoSuchJobError
from rq.job import Job as BaseJob
from rq.registry import BaseRegistry


def as_int(value) -> Optional[int]:
    if value is None:
        return
    return int(value)


def as_float(value) -> Optional[float]:
    if value is None:
        return
    return float(value)


def get_existing_job(job_id):
    """
    Returns a job identified by its ID, None otherwise
    """
    try:
        job = Job.fetch(job_id, connection=get_connection())
    except NoSuchJobError:
        return None
    return job


class Job(BaseJob):
    """
    Extension of RQ jobs to provide description updates and completion percentage
    """

    def __init__(self, *args, user_id: Optional[int] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._progress = None
        self.user_id = user_id

    @property
    def progress(self):
        return self.get_progress(refresh=False)

    def get_progress(self, refresh: bool = True):
        if refresh:
            self._progress = as_float(self.connection.hget(self.key, 'progress'))

        return self._progress

    def set_progress(self, progress: float):
        progress = as_float(progress)
        assert progress is not None and 0.0 <= progress <= 1.0, 'Progress should be a float between 0 and 1'
        self._progress = progress
        self.connection.hset(self.key, 'progress', self._progress)

    def set_description(self, description: Optional[str]):
        self.description = as_text(description)
        self.connection.hset(self.key, 'description', self.description)

    def to_dict(self, *args, **kwargs):
        """
        Serialize the job into a dict for Redis storage
        """
        obj = super().to_dict(*args, **kwargs)
        # Never include None values as those are not accepted by Redis
        if self._progress is not None:
            obj['progress'] = self._progress
        if self.user_id is not None:
            obj['user_id'] = self.user_id
        return obj

    def restore(self, raw_data):
        """
        Update job attributes from the Redis hash
        """
        super().restore(raw_data)
        obj = decode_redis_hash(raw_data)
        self._progress = as_float(obj.get('progress'))
        self.user_id = as_int(obj.get('user_id'))

    def delete(self, pipeline=None, remove_from_queue=True, **kwargs):
        """
        Overrides Job.delete, which already removes the job from all of RQ's registries and the queue
        before removing the job itself, to also remove the job from the UserRegistry when it has a user ID.

        Only remove when `remove_from_queue` is True, as users would otherwise 'lose' jobs as soon as they finish.
        """
        if remove_from_queue and self.user_id:
            registry = UserRegistry(
                self.origin,
                connection=self.connection,
                job_class=self.__class__,
                user_id=self.user_id,
            )
            registry.remove(self, pipeline=pipeline)

        super().delete(pipeline=pipeline, remove_from_queue=remove_from_queue, **kwargs)


class UserRegistry(BaseRegistry):
    """Job registry to index jobs per user_id
    The user_id value is extracted by the Queue class below
    when a job is enqueued"""

    def __init__(self, *args, user_id=None, **kwargs):
        # Build one sorted set per user in Redis
        super().__init__(*args, **kwargs)
        self.key = self.key_template.format(f'user:{user_id}')

    def cleanup(self):
        """
        This method is only here to prevent errors because this method is
        automatically called by `count()` and `get_job_ids()` methods
        implemented in BaseRegistry.
        An actual cleanup is implemented in the Job.delete method.
        """
        pass


class Queue(DjangoRQ):
    def user_registry(self, user_id):
        """Build a UserRegistry listing jobs for a specific user"""
        return UserRegistry(user_id=user_id, queue=self, job_class=self.job_class)

    def create_job(self, *args, **kwargs):
        # Extract user_id from delay() kwargs
        job_kwargs = kwargs.get('kwargs', {})
        user_id = job_kwargs.pop('user_id', None)
        description = job_kwargs.pop('description', None)

        # Build normal job
        job = super().create_job(*args, **kwargs)

        # Add the user ID to the job
        job.user_id = user_id

        # Add the description too, because .delay() does not allow a custom description
        # and instead uses the one defined in @job(description=…)
        if description:
            job.description = description

        # Add job to user registry
        if user_id is not None:
            reg = self.user_registry(user_id)
            reg.add(job)

        return job
