from collections.abc import Iterable, Iterator, Sized

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from django.db.models import Aggregate, CharField, Func
from django.db.models.expressions import BaseExpression, OrderByList
from django.urls import reverse

from arkindex.documents.models import Element, ElementPath


def build_absolute_url(element, request, name, id_argument='pk', **kwargs):
    """
    Build an absolute URL for a specified view using the element ID. Used by IIIF serializers.
    """
    kwargs[id_argument] = str(element.id)
    return request.build_absolute_uri(reverse(name, kwargs=kwargs))


def build_tree(tree, *, corpus, type):
    """
    Build Element and ElementPath instances from a tree described by
    a dict associating elements with their direct ancestors.
    Should only be used for unit tests.

    {'child': ['mother', 'father'], 'grandchild': 'child'}

    Returns a dict associating element names with created Elements.
    """
    assert isinstance(tree, dict)

    def parse_value(val):
        if isinstance(val, str):
            return [val]
        return val

    # List all elements
    # Pick all the keys, then add elements in the tree's values
    # Removes the need to add 'element': [] for root elements
    names = set(tree.keys()).union(*[parse_value(val) for val in tree.values()])
    # Create elements
    elements = {
        name: Element.objects.create(
            corpus=corpus,
            name=name,
            type=type,
        )
        for name in names
    }
    # Create empty paths for every element; add_parent will handle removing them if needed.
    ElementPath.objects.bulk_create([
        ElementPath(element=element, path=[], ordering=0)
        for element in elements.values()
    ])
    # Create paths
    for child_name, value in tree.items():
        parent_names = parse_value(value)
        for parent_name in parent_names:
            elements[child_name].add_parent(elements[parent_name])

    return elements


def bounding_box(linestring):
    """
    Returns a 4-tuple (x, y, width, height) for the bounding box of a LineString or LinearRing
    """
    all_x, all_y = zip(*linestring.coords)
    x, y = min(all_x), min(all_y)
    width, height = max(all_x) - x, max(all_y) - y
    return int(x), int(y), int(width), int(height)


def polygon_outside_image(img, linestring):
    x, y, w, h = bounding_box(linestring)
    return (x > img.width or x + w > img.width or y > img.height or y + h > img.height)


class BulkMap(Sized, Iterable):
    """
    Apply a function on the full list of items as soon as an iterator is requested.

    Useful to perform complex operations on a QuerySet with pagination
    without breaking through DRF's and Django's layers.
    """

    def __init__(self, func, iterable):
        assert callable(func), 'func must be callable'
        assert isinstance(iterable, Iterable), 'iterable must implement __iter__'
        assert isinstance(iterable, Sized), 'iterable must implement __len__'
        self.func = func
        self.iterable = iterable
        self._cached_length = None

    def __getitem__(self, key):
        assert isinstance(key, slice), 'BulkMap only supports slicing'
        return BulkMap(self.func, self.iterable[key])

    def __len__(self):
        if self._cached_length is None:
            # Proxy __len__ without calling the function
            if hasattr(self.iterable, 'count'):
                self._cached_length = self.iterable.count()
            else:
                self._cached_length = len(self.iterable)
        return self._cached_length

    def __iter__(self):
        return iter(self.func(list(self.iterable)))


class CounterIterator(Sized, Iterator):
    """
    Iterate over any iterator while keeping count of how many items were iterated on.
    This allows accessing the total amount of items in an iterator once a full iteration is completed
    without storing any item in memory.
    """

    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.count = 0

    def __next__(self):
        # This can fail with StopIteration if the iteration ends.
        item = next(self.iterator)
        # It didn't fail, so we can increment the counter.
        self.count += 1
        return item

    def __len__(self):
        return self.count


class RTrimChr(Func):
    """
    Django's RTrim function only allows trimming spaces.
    This function allows trimming any character, not only spaces.
    """
    function = 'RTRIM'
    arity = 2
    output_field = CharField()


class OrderedSetAggregate(Aggregate):
    """
    A subtype of aggregate functions provided by PostgreSQL, following the syntax
    `function (arguments) WITHIN GROUP (ORDER BY something)`.
    https://www.postgresql.org/docs/14/functions-aggregate.html#FUNCTIONS-ORDEREDSET-TABLE
    """
    template = '%(function)s(%(expressions)s) WITHIN GROUP (%(order_by)s)'
    allow_distinct = False

    def __init__(self, *expressions, order_by=None, **kwargs):
        # Handle order_by just like Django's `Window`
        # https://github.com/django/django/blob/4.1.3/django/db/models/expressions.py#L1654
        if isinstance(order_by, (list, tuple)):
            self.order_by = OrderByList(*order_by)
        elif isinstance(order_by, (BaseExpression, str)):
            self.order_by = OrderByList(order_by)
        else:
            raise TypeError('The `order_by=` keyword argument must be set to a valid `ORDER BY` clause.')
        super().__init__(*expressions, **kwargs)

    def resolve_expression(self, *args, **kwargs):
        resolved = super().resolve_expression(*args, **kwargs)
        # Also let the OrderByList resolve its expressions against the current query,
        # so that it can for example map a F('value') to the `value` column of a specific model
        resolved.order_by = self.order_by.resolve_expression(*args, **kwargs)
        return resolved

    def as_sql(self, compiler, connection, **extra_context):
        # Let the OrderByList turn itself into a valid ORDER BY clause
        order_by_sql, order_by_params = self.order_by.as_sql(compiler, connection)
        # Use that ORDER BY into our own template
        sql, params = super().as_sql(compiler, connection, order_by=order_by_sql, **extra_context)
        return sql, (*params, *order_by_params)


class PercentileCont(OrderedSetAggregate):
    """
    Continuous percentile: the value matching a specified fraction within an ordered set of values.
    Unlike a discrete percentile, this will interpolate between two values if needed.
    """
    function = 'percentile_cont'


def build_public_key() -> str:
    """
    Quickly build a Ponos public key as a PEM-encoded string
    """
    return (
        ec.generate_private_key(
            ec.SECP384R1(),
            default_backend(),
        )
        .public_key()
        .public_bytes(
            Encoding.PEM,
            PublicFormat.SubjectPublicKeyInfo,
        )
        .decode("utf-8")
    )
