from rest_framework import serializers


class XorValidator(object):
    """
    A generic validator for when two fields can't be used simultaneously
    """
    requires_context = True

    def __init__(self, field1, field2, required=True):
        self.field1 = field1
        self.field2 = field2
        self.required = required

    def __call__(self, attrs, serializer):
        # This only checks the fields being sent in the request, not the ones that might already have a value
        # on the instance if it is an update / partial update request being sent.
        if attrs.get(self.field1) is not None and attrs.get(self.field2) is not None:
            raise serializers.ValidationError(f'Only one of `{self.field1}` and `{self.field2}` may be set.')
        if attrs.get(self.field1) is None and attrs.get(self.field2) is None and self.required is True:
            if serializer.partial:
                return
            raise serializers.ValidationError(f'One of `{self.field1}` or `{self.field2}` must be set.')


class ConditionalUniqueValidator(serializers.UniqueTogetherValidator):
    """
    A UniqueTogetherValidator that only runs when a condition is fulfilled.
    This is useful to validate partial unique constraints, which use `condition=Q(…)`.

    The `condition` keyword argument should be a function that takes a dict to validate
    and returns True if the unique constraint should be checked and False otherwise.
    The dict is the same as found in `ModelSerializer.validate(value)` as the `value`.
    """

    def __init__(self, *args, condition=None, **kwargs):
        assert callable(condition), 'The condition should be a function.'
        self.condition = condition
        super().__init__(*args, **kwargs)

    def __call__(self, attrs, serializer):
        if not self.condition(attrs):
            return
        return super().__call__(attrs, serializer)


class ForbiddenValidator(object):
    """
    A validator that will show an error message any time a value is set.

    Note that using this validator on a field that is `read_only` is useless,
    and using it on a `required` field will make the serializer always fail.

    Sending a null value for the field will not cause this validator to run:
    with `allow_null=True`, the value gets ignored, and with `allow_null=False`,
    DRF will cause its default "This field may not be null".
    """

    def __init__(self, message='This field is forbidden.'):
        self.message = message

    def __call__(self, data):
        if data is not serializers.empty:
            raise serializers.ValidationError(self.message)
